#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

from selinon import SystemState
from selinon.selective import compute_selective_run
from selinon_test_case import SelinonTestCase


class TestSelectiveFlow(SelinonTestCase):
    def get_initial_system_state(self, flow_name, node_args, task_names, follow_subflows=False, run_subsequent=False):
        """ Get SystemState instance for the given selective flow

        :param flow_name: name of the flow to get SystemState instance for
        :param node_args: arguments that are passed to the flow
        :param task_names: a list of tasks that should be selectively run
        :param follow_subflows: if True, subflows will be inspected for task_names to be run
        :param run_subsequent: run all nodes that are subsequent nodes of the tasks_names nodes
        """
        selective = compute_selective_run(flow_name, task_names, follow_subflows, run_subsequent)
        return SystemState(id(self), 'flow1', node_args, selective=selective)

    def test_selective_flow_first_task(self):
        #
        # flow1:
        #
        #     Task1    Task2    Task3
        #                |        |
        #              Task4    Task5
        #
        # Note: Only Task2 should be run
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1', 'Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task3'], 'to': ['Task5'], 'condition': self.cond_true}],
        }
        node_args = {'foo': 'bar'}
        self.init(edge_table)

        system_state = self.get_initial_system_state('flow1', node_args, ['Task2'],
                                                     follow_subflows=False, run_subsequent=False)

        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task2'} == set(self.instantiated_tasks)

        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task2'} == set(self.instantiated_tasks)

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2, None)

        assert task2.node_args == node_args

        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()

        assert retry is None

    def test_selective_flow_one_task(self):
        #
        # flow1:
        #
        #         Task1
        #           |  \
        #           |   \
        #         Task2  Task4
        #           |
        #           |
        #         Task3
        #
        # Note: Only Task2 and Task1 should be run
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true}],
        }
        node_args = {'foo': 'bar'}
        self.init(edge_table)

        system_state = self.get_initial_system_state('flow1', node_args, ['Task2'],
                                                     follow_subflows=False, run_subsequent=False)

        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1'} == set(self.instantiated_tasks)

        self.set_finished(self.get_task('Task1'), None)
        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2'} == set(self.instantiated_tasks)

        self.set_finished(self.get_task('Task2'), None)
        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        assert retry is None

    def test_selective_flow_two_tasks(self):
        #
        # flow1:
        #
        #         Task1        Task2        Task3        Task4     Task5
        #           \            /             \         /           |
        #            \          /               \       /          Task10
        #             ----------                 -------
        #                  |                     /    \
        #                Task6              Task7      Task8
        #                  |                  |
        #                  |                  |
        #                Task11             Task9
        #
        # Note: Only Task6 and Task7 should be run
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1', 'Task2', 'Task3', 'Task4', 'Task5'], 'condition': self.cond_true},
                      {'from': ['Task1', 'Task2'], 'to': ['Task6'], 'condition': self.cond_true},
                      {'from': ['Task3', 'Task4'], 'to': ['Task7', 'Task8'], 'condition': self.cond_true},
                      {'from': ['Task7'], 'to': ['Task9'], 'condition': self.cond_true},
                      {'from': ['Task5'], 'to': ['Task10'], 'condition': self.cond_true}],
        }
        self.init(edge_table)

        system_state = self.get_initial_system_state('flow1', None, ['Task6', 'Task7'],
                                                     follow_subflows=False, run_subsequent=False)

        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2', 'Task3', 'Task4'} == set(self.instantiated_tasks)

        # Task1 has finished
        self.set_finished(self.get_task('Task1'), None)

        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2', 'Task3', 'Task4'} == set(self.instantiated_tasks)

        # Task2 has finished
        self.set_finished(self.get_task('Task2'), None)

        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2', 'Task3', 'Task4', 'Task6'} == set(self.instantiated_tasks)

        # Task3 has finished
        self.set_finished(self.get_task('Task3'), None)

        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2', 'Task3', 'Task4', 'Task6'} == set(self.instantiated_tasks)

        # Task4 has finished
        self.set_finished(self.get_task('Task4'), None)

        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2', 'Task3', 'Task4', 'Task6', 'Task7'} == set(self.instantiated_tasks)

        # Task7 has finished
        self.set_finished(self.get_task('Task7'), None)

        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2', 'Task3', 'Task4', 'Task6', 'Task7'} == set(self.instantiated_tasks)

        # Task6 has finished
        task6 = self.get_task('Task6')
        self.set_finished(task6, None)

        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        assert retry is None

    def test_selective_flow_subflow(self):
        #
        # flow1:
        #
        #     flow2    Task1
        #        |       |
        #        |      Task2
        #        |       |  \
        #         -------    Task3
        #            |
        #            |
        #          Task4
        #            |
        #            |
        #          Task5
        #
        # flow2:
        #
        #     TaskF2
        #       |
        #       |
        #     Task4
        #       |  \
        #       |   \
        #     flow3  flow4
        #
        # flow3:
        #
        #    Task4
        #      |
        #      |
        #    TaskF3
        #
        # flow4:
        #
        #    TaskF3
        #
        # Note: Only Task4 should be run, flow4 is not executed
        #
        def simulate_flow2(flow2, f2_already_instantiated_tasks):
            f2_system_state = flow2.get_initial_system_state()

            f2_retry = f2_system_state.update()
            f2_state_dict = f2_system_state.to_dict()
            f2_selective = f2_system_state.selective

            assert f2_retry is not None
            assert f2_already_instantiated_tasks | {'TaskF2'} == set(self.instantiated_tasks)
            self.set_finished(self.get_task('TaskF2'))

            f2_system_state = SystemState(id(self), 'flow2', state=f2_state_dict,
                                          node_args=f2_system_state.node_args, selective=f2_selective)
            f2_retry = f2_system_state.update()
            f2_state_dict = f2_system_state.to_dict()
            f2_selective = f2_system_state.selective

            assert f2_retry is not None
            assert f2_already_instantiated_tasks | {'TaskF2', 'Task4'} == set(self.instantiated_tasks)
            assert 'flow3' not in self.instantiated_flows

            self.set_finished(self.get_task('Task4'))
            f2_system_state = SystemState(id(self), 'flow2', state=f2_state_dict,
                                          node_args=f2_system_state.node_args, selective=f2_selective)
            f2_retry = f2_system_state.update()
            f2_state_dict = f2_system_state.to_dict()
            f2_selective = f2_system_state.selective

            assert f2_retry is not None
            assert 'flow3' in self.instantiated_flows
            simulate_flow3(self.get_flow('flow3'), set(self.instantiated_tasks))

            f2_system_state = SystemState(id(self), 'flow2', state=f2_state_dict,
                                          node_args=f2_system_state.node_args, selective=f2_selective)
            f2_retry = f2_system_state.update()

            assert f2_retry is None
            self.set_finished(flow2, None)

            # let's clean after self so we inspect purely flow1 in callee
            self.remove_all_tasks_by_name('TaskF2')
            self.remove_all_tasks_by_name('Task4')
            self.remove_all_flows_by_name('flow2')

        def simulate_flow3(flow3, f3_already_instantiated_tasks):
            f3_system_state = flow3.get_initial_system_state()

            f3_retry = f3_system_state.update()
            f3_state_dict = f3_system_state.to_dict()
            f3_selective = f3_system_state.selective

            assert f3_retry is not None
            assert f3_already_instantiated_tasks | {'Task4'} == set(self.instantiated_tasks)
            self.set_finished(self.get_task('Task4'))

            f3_system_state = SystemState(id(self), 'flow3', state=f3_state_dict,
                                          node_args=f3_system_state.node_args, selective=f3_selective)
            f3_retry = f3_system_state.update()
            assert f3_retry is None

            self.set_finished(flow3, None)
            self.remove_all_flows_by_name('flow3')

        edge_table = {
            'flow1': [{'from': [], 'to': ['flow2', 'Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['flow2', 'Task2'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task4'], 'to': ['Task5'], 'condition': self.cond_true}],
            'flow2': [{'from': [], 'to': ['TaskF2'], 'condition': self.cond_true},
                      {'from': ['TaskF2'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task4'], 'to': ['flow4', 'flow3'], 'condition': self.cond_true}],
            'flow3': [{'from': [], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': ['Task4'], 'to': ['TaskF3'], 'condition': self.cond_true}],
            'flow4': [{'from': [], 'to': ['TaskF3']}]
        }
        self.init(edge_table)

        system_state = self.get_initial_system_state('flow1', None, ['Task4'],
                                                     follow_subflows=True, run_subsequent=False)

        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1'} == set(self.instantiated_tasks)
        assert {'flow2'} == set(self.instantiated_flows)

        simulate_flow2(self.get_flow('flow2'), set(self.instantiated_tasks))
        self.set_finished(self.get_task('Task1'), None)
        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2'} == set(self.instantiated_tasks)

        self.set_finished(self.get_task('Task2'), None)
        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2', 'Task4'} == set(self.instantiated_tasks)

        self.set_finished(self.get_task('Task4'), None)

        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        assert retry is None
        assert 'flow4' not in self.instantiated_flows

    def test_selective_flow_sybsequent_cyclic(self):
        #
        #      Task1  <-
        #       /  |    |
        #      /   |    |
        # Task3  Task2 --
        #
        # This is a very special case, where we don't want to keep state in dispatcher, so all tasks will be run
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = self.get_initial_system_state('flow1', None, ['Task2'],
                                                     follow_subflows=False, run_subsequent=True)

        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1'} == set(self.instantiated_tasks)

        self.set_finished(self.get_task('Task1'), None)
        system_state = SystemState(id(self), 'flow1',
                                   state=state_dict, node_args=system_state.node_args, selective=selective)
        retry = system_state.update()
        state_dict = system_state.to_dict()
        selective = system_state.selective

        assert retry is not None
        assert {'Task1', 'Task2', 'Task3'} == set(self.instantiated_tasks)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Main system actions done by Selinon."""

from collections import deque
import copy
import datetime
from functools import reduce
import itertools
import traceback

from .celery import AsyncResult
from .config import Config
from .errors import CacheMissError
from .errors import DispatcherRetry
from .errors import FlowError
from .errors import StorageError
from .lock_pool import LockPool
from .selective import compute_selective_run
from .storage_pool import StoragePool
from .task_envelope import SelinonTaskEnvelope
from .trace import Trace


class SystemState:  # pylint: disable=too-many-instance-attributes
    """Main system actions done by Selinon."""

    # Throttled nodes in the current worker: node name -> next schedule time
    # Throttle should be safe with concurrency
    _throttle_lock_pool = LockPool()
    _throttled_tasks = {}
    _throttled_flows = {}
    _node_state_cache_lock = LockPool()

    @property
    def node_args(self):
        """Node arguments.

        :return: arguments for a node
        """
        return self._node_args

    def _get_async_result(self, node_name, node_id):  # pylint: disable=invalid-name,redefined-builtin
        """Retrieve async result, check cache first.

        :param node_name: a name of node for which async result should be checked
        :param node_id: id if node for which async result should be checked
        :return: Celery AsyncResult
        """
        cache = Config.async_result_cache[self._flow_name]
        trace_msg = {
            'flow_name': self._flow_name,
            'node_args': self._node_args,
            'parent': self._parent,
            'dispatcher_id': self._dispatcher_id,
            'queue': Config.dispatcher_queues[self._flow_name],
            'node_id': node_id,
            'node_name': node_name,
            'selective': self._selective
        }

        with self._node_state_cache_lock.get_lock(self._flow_name):
            res = None
            result_retrieved_from_cache = False

            Trace.log(Trace.NODE_STATE_CACHE_GET, trace_msg)
            try:
                res = cache.get(node_id)
                result_retrieved_from_cache = True
            except CacheMissError:
                Trace.log(Trace.NODE_STATE_CACHE_MISS, trace_msg, what=traceback.format_exc())
            except Exception:  # pylint: disable=broad-except
                Trace.log(Trace.NODE_STATE_CACHE_ISSUE, trace_msg, what=traceback.format_exc())
            else:
                Trace.log(Trace.NODE_STATE_CACHE_HIT, trace_msg)

            if not result_retrieved_from_cache:
                try:
                    res = AsyncResult(id=node_id)
                    successful = res.successful()
                    failed = res.failed()
                except Exception as exc:  # pylint: disable=broad-except
                    Trace.log(Trace.RESULT_BACKEND_ISSUE, trace_msg, what=traceback.format_exc())
                    raise DispatcherRetry(keep_state=True, adjust_retry_count=False) from exc

                # We can cache only results of tasks that have finished or failed, not the ones that are
                # going to be processed (state will change).
                if successful or failed:
                    Trace.log(Trace.NODE_STATE_CACHE_ADD, trace_msg)
                    try:
                        cache.add(node_id, res)
                    except Exception:  # pylint: disable=broad-except
                        Trace.log(Trace.NODE_STATE_CACHE_ISSUE, trace_msg, what=traceback.format_exc())

            return res

    def _instantiate_active_nodes(self, arr):
        """Retrieve all async results for active nodes.

        :return: convert node references from argument to AsyncResult
        """
        return [{'name': node['name'], 'id': node['id'],
                 'result': self._get_async_result(node['name'], node['id'])} for node in arr]

    @staticmethod
    def _deinstantiate_active_nodes(arr):
        """Store only id's of AsyncResults.

        Used before serialization so we get JSON.

        :param arr: array of nodes that are active in a flow
        :return: node references for Dispatcher arguments
        """
        return [{'name': x['name'], 'id': x['id']} for x in arr]

    @staticmethod
    def _idxs2items(edge_node_table, arr):
        """For the given array, get all edges based on edge node table.

        :param edge_node_table: edge table to be used
        :param arr: a list of indexes to edge table
        :return: edges that correspond to the given index array
        """
        # we keep only the index to edges_table in order to avoid serialization and optimize number representation
        ret = []
        for i in arr:
            ret.append(edge_node_table[i])
        return ret

    def __repr__(self):  # noqa
        # Make tests more readable
        return str(self.to_dict())

    def __init__(self, dispatcher_id, flow_name, node_args=None, retry=None, state=None, parent=None, selective=None):
        # pylint: disable=too-many-arguments
        """Instantiate system state computation (called from Dispatcher).

        :param dispatcher_id: id of dispatcher for the current flow
        :param flow_name: name of the flow
        :param node_args: flow arguments
        :param retry: retry information
        :param state: current state (serialized, if any)
        :param parent: information about parent nodes
        :param selective: precomputed information about selective flow, if any
        """
        state_dict = state or {}

        self._dispatcher_id = dispatcher_id
        self._flow_name = flow_name
        self._node_args = node_args
        self._parent = parent or {}
        self._selective = selective or False
        self._active_nodes = self._instantiate_active_nodes(state_dict.get('active_nodes', []))
        self._finished_nodes = state_dict.get('finished_nodes', {})
        self._failed_nodes = state_dict.get('failed_nodes', {})
        self._waiting_edges_idx = state_dict.get('waiting_edges', [])
        self._triggered_edges_idx = set(state_dict.get('triggered_edges', {}))
        # Instantiate lazily later if we will know that there is something to process
        self._waiting_edges = []
        self._retry = retry

        # TODO: fix this - for some reasons serializer uses strings in keys
        if self._selective:
            for flow in self._selective['waiting_edges_subset']:
                self._selective['waiting_edges_subset'][flow] = \
                    {int(k): v for k, v in self._selective['waiting_edges_subset'][flow].items()}

    def to_dict(self):
        """Serialize current system state.

        :return: converted system state to dict
        """
        return {
            'active_nodes': self._deinstantiate_active_nodes(self._active_nodes),
            'finished_nodes': self._finished_nodes,
            'failed_nodes': self._failed_nodes,
            'waiting_edges': self._waiting_edges_idx,
            # Convert to a list due to JSON serialization
            'triggered_edges': list(self._triggered_edges_idx)
        }

    @property
    def selective(self):
        """All edges that should be started selectively as computed by compute_selective."""
        return self._selective

    def _get_countdown(self, node_name, is_flow):
        """Get countdown for throttling.

        :param node_name: node name
        :param is_flow: true if node_name is a flow
        :return: countdown seconds for the current schedule
        """
        if is_flow:
            throttle_conf = Config.throttle_flows
            throttled_nodes = self._throttled_flows
        else:
            throttle_conf = Config.throttle_tasks
            throttled_nodes = self._throttled_tasks

        with self._throttle_lock_pool.get_lock(node_name):
            if throttle_conf[node_name]:
                current_datetime = datetime.datetime.now()
                if node_name not in throttled_nodes:
                    # we throttle for the first time
                    throttled_nodes[node_name] = current_datetime
                    return None

                next_run = current_datetime + throttle_conf[node_name]
                countdown = (throttled_nodes[node_name] + throttle_conf[node_name] - current_datetime).total_seconds()
                throttled_nodes[node_name] = next_run

                return countdown if countdown > 0 else None

        return None

    def _get_successful_and_failed(self, reused):
        """Retrieve info about successful and failed nodes.

        :return: all successful and failed nodes in system from active nodes
        """
        ret_successful = reused
        ret_failed = []

        trace_msg = {
            'flow_name': self._flow_name,
            'dispatcher_id': self._dispatcher_id,
            'selective': self._selective
        }

        new_active_nodes = []
        for node in self._active_nodes:
            if node['result'].successful():
                Trace.log(Trace.NODE_SUCCESSFUL, trace_msg, node_name=node['name'], node_id=node['id'])
                ret_successful.append(node)
            elif node['result'].failed():
                Trace.log(
                    Trace.NODE_FAILURE,
                    trace_msg,
                    node_name=node['name'],
                    node_id=node['id'],
                    what=node['result'].traceback
                )
                # We keep track of failed nodes to handle failures once all nodes finish
                if node['name'] not in self._failed_nodes:
                    self._failed_nodes[node['name']] = []
                self._failed_nodes[node['name']].append(node['id'])

                # Retry/Fail if this node was marked as eager failure node
                eager = Config.eager_failures.get(self._flow_name, False)
                if eager is True or (isinstance(eager, list) and node['name'] in eager):
                    Trace.log(
                        Trace.EAGER_FAILURE,
                        trace_msg,
                        node_name=node['name'],
                        node_id=node['id'],
                        what=node['result'].traceback
                    )
                    self._active_nodes.remove(node)
                    raise FlowError(self.to_dict())

                ret_failed.append(node['id'])
            else:
                new_active_nodes.append(node)

        self._active_nodes = new_active_nodes
        return ret_successful, ret_failed

    def _execute_selective_run_func(self, node_name, node_args, parent):
        """Execute selective run function so we know whether we need to re-run the given node.

        :param node_name: name of the node for which the selective function should be executed
        :param node_args: arguments that would be passed to desired node
        :param parent: information about parent nodes
        :return: None if node should be run, id of task that result's should be reused
        """
        trace_msg = {
            'flow_name': self._flow_name,
            'dispatcher_id': self._dispatcher_id,
            'selective': self._selective,
            'node_name': node_name,
            'node_args': node_args,
            'parent': parent,
        }

        storage_pool = StoragePool(parent, self._flow_name)
        selective_func = Config.selective_run_task[node_name]

        result = selective_func(self._flow_name, node_name, node_args, self._selective['task_names'], storage_pool)
        Trace.log(Trace.SELECTIVE_RUN_FUNC, trace_msg, {'result': result})

        return result

    def _start_node(self, node_name, parent, node_args, edge=None, force_propagate_node_args=False):
        """Start a node in the system.

        :param node_name: name of a node to be started
        :param parent: parent nodes of the starting node
        :param node_args: arguments for the starting node
        :param edge: edge that triggered node start, can be None for fallbacks
        """
        # pylint: disable=too-many-arguments
        from .dispatcher import Dispatcher

        if Config.is_flow(node_name):
            start_node_args = None
            if force_propagate_node_args or Config.should_propagate_node_args(self._flow_name, node_name):
                start_node_args = node_args

            start_parent = None
            if Config.should_propagate_parent(self._flow_name, node_name):
                start_parent = parent

            selective = None
            if edge and edge.get('selective'):
                selective = compute_selective_run(node_name, **edge['selective'])

            kwargs = {
                'flow_name': node_name,
                'node_args': start_node_args,
                'parent': start_parent,
                'selective': selective or self.selective
            }

            countdown = self._get_countdown(node_name, is_flow=True)
            async_result = Dispatcher().apply_async(  # pylint: disable=assignment-from-no-return
                kwargs=kwargs,
                queue=Config.dispatcher_queues[node_name],
                countdown=countdown
            )

            Trace.log(Trace.SUBFLOW_SCHEDULE, {
                'flow_name': self._flow_name,
                'condition_str': None if not edge else edge['condition_str'],
                'foreach_str': None if not edge else edge.get('foreach_str'),
                'selective_edge_conf': None if not edge else edge.get('selective', False),
                'child_flow_name': node_name,
                'dispatcher_id': self._dispatcher_id,
                'child_dispatcher_id': async_result.task_id,
                'queue': Config.dispatcher_queues[node_name],
                'countdown': countdown,
                'child_selective': selective,
                'selective': self._selective,
                'node_args': start_node_args
            })

        else:
            kwargs = {
                'task_name': node_name,
                'flow_name': self._flow_name,
                'parent': parent,
                'node_args': node_args,
                'dispatcher_id': self._dispatcher_id
            }

            countdown = self._get_countdown(node_name, is_flow=False)
            async_result = SelinonTaskEnvelope().apply_async(  # pylint: disable=assignment-from-no-return
                kwargs=kwargs,
                queue=Config.task_queues[node_name],
                countdown=countdown
            )

            Trace.log(Trace.TASK_SCHEDULE, kwargs, {
                'task_id': async_result.task_id,
                'queue': Config.task_queues[node_name],
                'condition_str': None if not edge else edge['condition_str'],
                'foreach_str': None if not edge else edge.get('foreach_str'),
                'selective_edge': False,  # always False as we are starting a task
                'countdown': countdown,
                'selective': self._selective
            })

        record = {
            'name': node_name,
            'id': async_result.task_id,
            'result': async_result
        }

        if node_name not in Config.nowait_nodes.get(self._flow_name, []):
            self._active_nodes.append(record)

        return record

    def _fire_edge(self, edge_idx, edge, storage_pool, parent, node_args):
        """Fire edge - start new nodes as described in edge table.

        :param edge: edge that should be fired
        :param storage_pool: storage pool which makes results of previous tasks available
        :param parent: parent nodes
        :param node_args: node arguments
        :return: list of nodes that were scheduled
        """
        # pylint: disable=too-many-arguments,too-many-branches,too-many-locals
        started = []
        selective_reuse = []
        trace_msg = {
            'nodes_to': edge['to'],
            'nodes_from': edge['from'],
            'flow_name': self._flow_name,
            'foreach_str': edge.get('foreach_str'),
            'condition_str': edge['condition_str'],
            'parent': parent,
            'node_args': self._node_args,
            'dispatcher_id': self._dispatcher_id,
            'selective': self._selective
        }

        nodes2start = None
        if self._selective:
            if edge_idx not in self._selective['waiting_edges_subset'][self._flow_name].keys():
                Trace.log(Trace.SELECTIVE_OMIT_EDGE, trace_msg)
                return [], []

            nodes2start = self._selective['waiting_edges_subset'][self._flow_name][edge_idx]

        if 'foreach' in edge:
            iterable = edge['foreach'](storage_pool, node_args)
            Trace.log(Trace.FOREACH_RESULT, trace_msg, {'result': iterable})
            # handle None as well
            if not iterable:
                return started, []
            for res in iterable:
                for node_name in edge['to']:
                    if nodes2start and node_name not in nodes2start:
                        Trace.log(Trace.SELECTIVE_OMIT_NODE, trace_msg, {'omitted_node': node_name})
                        continue

                    if self._selective and node_name not in self._selective['task_names'] \
                            and not Config.is_flow(node_name):
                        selective_run_func_result = self._execute_selective_run_func(node_name, res, parent)
                        if selective_run_func_result:
                            Trace.log(Trace.SELECTIVE_TASK_REUSE, trace_msg, {
                                'task_name': node_name,
                                'task_id': selective_run_func_result
                            })
                            selective_reuse.append({
                                'name': node_name,
                                'id': selective_run_func_result,
                                'result': self._get_async_result(node_name, selective_run_func_result)
                            })
                            continue

                    if edge.get('foreach_propagate_result'):
                        record = self._start_node(node_name, parent, res, edge, force_propagate_node_args=True)
                    else:
                        record = self._start_node(node_name, parent, node_args, edge)
                    started.append(record)
        else:
            for node_name in edge['to']:
                if nodes2start and node_name not in nodes2start:
                    Trace.log(Trace.SELECTIVE_OMIT_NODE, trace_msg, {'omitted_node': node_name})
                    continue

                if self._selective and node_name not in self._selective['task_names'] and not Config.is_flow(node_name):
                    selective_run_func_result = self._execute_selective_run_func(node_name, node_args, parent)
                    if selective_run_func_result:
                        Trace.log(Trace.SELECTIVE_TASK_REUSE, trace_msg, {
                            'task_name': node_name,
                            'task_id': selective_run_func_result
                        })
                        selective_reuse.append({
                            'name': node_name,
                            'id': selective_run_func_result,
                            'result': self._get_async_result(node_name, selective_run_func_result)
                        })
                        continue

                record = self._start_node(node_name, parent, node_args, edge)
                started.append(record)

        return started, selective_reuse

    def _run_fallback(self, failure_node, combination):
        """Evaluate fallback condition and run fallback iff condition is true.

        :param failure_node: failure node that should be evaluated
        :param combination: combination of failed nodes
        :return: triplet started - records about started nodes, should_continue - True if other fallbacks can be
                 run, skip_failure_node - true if the current failure node should be skipped in this fallback run
                 (preserves infinite deps)
        """
        traced_nodes_arr = []
        # compute affected nodes
        for node in combination:
            if node[0] not in self._failed_nodes.keys():
                # this means that some fallback was previously run and some node in the combination was already handled
                return [], True, True
            traced_nodes_arr.append({'name': node[0], 'id': self._failed_nodes[node[0]][0]})

        trace_dict = {
            'flow_name': self._flow_name,
            'dispatcher_id': self._dispatcher_id,
            'nodes': traced_nodes_arr,
            'selective': self._selective,
            'fallback': None
        }

        should_continue = True
        skip_failure_node = True
        started = []
        affected_nodes = set()

        for idx, fallback in enumerate(failure_node['fallback']):
            trace_dict['fallback'] = fallback
            trace_dict['condition_strs'] = failure_node['condition_strs'][idx]
            if not failure_node['conditions'][idx](StoragePool({}, self._flow_name), self._node_args):
                Trace.log(Trace.FALLBACK_COND_FALSE, trace_dict)
                continue

            # preserve infinite loops when no fallback is run due to condition evaluation - proceed to next failure node
            skip_failure_node = False
            Trace.log(Trace.FALLBACK_COND_TRUE, trace_dict)

            # we will run some fallback, remove affected failed nodes from failed_nodes
            for node in combination:
                affected_nodes.add(node[0])

            Trace.log(Trace.FALLBACK_START, trace_dict)
            if fallback is True:
                continue

            for node_name in fallback:
                record = self._start_node(node_name, parent=None, node_args=self._node_args)
                if node_name not in Config.nowait_nodes[self._flow_name]:
                    started.append(record)
                else:
                    # wait for fallback to finish in order to avoid time dependent flow evaluation
                    should_continue = False

        # clean nodes for which a fallback was run
        for node_name in affected_nodes:
            self._failed_nodes[node_name].pop(0)
            if not self._failed_nodes[node_name]:
                del self._failed_nodes[node_name]

        return started, should_continue, skip_failure_node

    def _compute_and_run_fallback(self):
        """Run fallback in the system.

        :return: fallbacks that were run
        """
        # we sort it first to make evaluation dependent on alphabetical order
        ret = []
        failed_nodes = sorted(self._failed_nodes.items())

        # pylint: disable=too-many-nested-blocks
        for i in range(len(failed_nodes), 0, -1):
            for combination in itertools.combinations(failed_nodes, i):
                try:
                    failure_nodes = Config.failures[self._flow_name]
                    failure_node = reduce(lambda n, c: n['next'][c[0]], combination[1:],
                                          failure_nodes[combination[0][0]])
                except KeyError:
                    # such failure not found in the tree of permutations - this means that this
                    # flow will always fail, but run defined fallbacks for defined failures first
                    continue

                while True:
                    fallback_run, should_continue, skip_failure_node = self._run_fallback(failure_node, combination)
                    ret.extend(fallback_run)

                    if not should_continue:
                        return ret

                    if skip_failure_node or not all(node[0] in self._failed_nodes.keys() for node in combination):
                        break

        return ret

    def _update_waiting_edges(self, nodes):
        """Update waiting edges (edges that wait for a node finish) based on nodes that finished.

        :param nodes: nodes that will trigger edges.
        """
        res = []

        for node in nodes:
            for idx, edge in enumerate(Config.edge_table[self._flow_name]):
                if node['name'] not in edge['from']:
                    continue

                if idx in self._waiting_edges_idx:
                    continue

                if node['name'] in Config.nowait_nodes.get(self._flow_name, []):
                    continue

                if self._selective and idx not in self._selective['waiting_edges_subset'][self._flow_name]:
                    continue

                res.append(node)
                self._waiting_edges.append(edge)
                self._waiting_edges_idx.append(idx)

        return res

    def _extend_parent_from_flow(self, dst_dict, flow_name, flow_id, key, compound=False):
        # pylint: disable=too-many-arguments,too-many-locals
        """Compute parent in a flow in case of propagate_parent flag.

        :param dst_dict: a dictionary that should be extended with calculated parents
        :param flow_name: a flow name for which propagate_parent is calculated
        :param flow_id: flow identifier as identified in Celery
        :param key: key under which should be result stored
        :param compound: true if propagate_compound was used
        """
        def follow_keys(d, k):
            # pylint: disable=invalid-name,missing-docstring
            if not compound:
                # A key is always a flow name except the last key, which is a task name with list of task ids
                for m_k in k:
                    if m_k not in d:
                        d[m_k] = {}
                    d = d[m_k]
                return d

            # If we have compound, the key to list of task is flow name.
            # This will allow to use conditions as expected.
            # e.g.:
            #  {'flow1': {'Task1': ['task1_id1', 'task1_id2'],
            #            {'Task2': ['task2_id1', 'task2_id2']}
            if flow_name not in d:
                d[flow_name] = {}
            return d[flow_name]

        def push_flow(s, p_n_name, flow_result, k):
            # pylint: disable=invalid-name,missing-docstring
            for n_name, n_ids in flow_result[key].items():
                if not compound:
                    shallow_k = copy.copy(k)
                    shallow_k.append(p_n_name)
                    s.append((n_name, n_ids, shallow_k))
                else:
                    s.append((n_name, n_ids, k))

        res = dst_dict

        async_result = self._get_async_result(flow_name, flow_id)
        stack = deque()
        push_flow(stack, flow_name, async_result.result, deque())

        while stack:
            node_name, node_ids, keys = stack.pop()
            if Config.is_flow(node_name):
                for node_id in node_ids:
                    push_flow(stack, node_name, self._get_async_result(node_name, node_id).result, keys)
            else:
                dst = follow_keys(res, keys)

                if node_name not in dst:
                    dst[node_name] = []

                for node_id in node_ids:
                    dst[node_name].append(node_id)

        return res

    def _start_new_from_finished(self, new_finished):  # pylint: disable=too-many-locals
        """Start new based on finished nodes.

        :param new_finished: finished nodes based on which we should start new nodes
        :return: newly started nodes
        """
        # pylint: disable=too-many-nested-blocks,too-many-branches
        new_started_nodes = []
        selective_reuse = []

        if len(new_finished) == 1 and not self._active_nodes and not self._finished_nodes:
            # propagate arguments from newly finished node if configured to do so
            if Config.node_args_from_first.get(self._flow_name, False):
                self._node_args = StoragePool.retrieve(self._flow_name, new_finished[0]['name'], new_finished[0]['id'])

        for node in new_finished:
            # We could optimize this by pre-computing affected edges in pre-generated config file for each
            # node and computing intersection with waiting edges, but let's stick with this solution for now
            edges = []
            for waiting_edges_idx, edge_table_idx in enumerate(self._waiting_edges_idx):
                # we have to store index to edge table, but inspect possible edge fires only in waiting edges
                if node['name'] in self._waiting_edges[waiting_edges_idx]['from']:
                    edges.append((edge_table_idx, self._waiting_edges[waiting_edges_idx]))

            for i, edge in edges:
                from_nodes = dict.fromkeys(edge['from'], [])

                for from_name in from_nodes:
                    if from_name == node['name']:
                        from_nodes[from_name] = [{'name': node['name'], 'id': node['id']}]
                    else:
                        from_nodes[from_name] = [{'name': from_name, 'id': n}
                                                 for n in self._finished_nodes.get(from_name, [])]

                # if there are multiple nodes of a same type waiting on an edge:
                #
                #   A    B
                #    \  /
                #     \/
                #     C
                #
                # A: <id1>, <id2>
                # B: <id3>, <id4>
                #
                # Compute all combinations: (<id1>, <id3>), (<id1>, <id4>), (<id2>, <id3>), (<id2>, <id4>)
                #
                for start_nodes in itertools.product(*from_nodes.values()):
                    parent = {}
                    storage_id_mapping = {}

                    for start_node in start_nodes:
                        if Config.is_flow(start_node['name']):
                            compound = Config.should_propagate_compound_finished(self._flow_name,
                                                                                 start_node['name'])
                            if Config.should_propagate_finished(self._flow_name, start_node['name']) or compound:
                                parent = self._extend_parent_from_flow(parent, start_node['name'], start_node['id'],
                                                                       'finished_nodes', compound)
                        else:
                            parent[start_node['name']] = start_node['id']
                            storage_id_mapping[start_node['name']] = start_node['id']

                    # We could also examine results of subflow, there could be passed a list of subflows with
                    # finished_nodes to 'condition' in order to do inspection
                    storage_pool = StoragePool(storage_id_mapping, self._flow_name)

                    try:
                        condition_result = edge['condition'](storage_pool, self._node_args)
                    except StorageError as exc:
                        Trace.log(Trace.STORAGE_ISSUE, what=traceback.format_exc())
                        raise DispatcherRetry(keep_state=True, adjust_retry_count=False) from exc

                    if condition_result:
                        records, reused = self._fire_edge(i, edge, storage_pool,
                                                          parent=parent, node_args=self._node_args)
                        new_started_nodes.extend(records)
                        self._triggered_edges_idx.add(i)
                        selective_reuse.extend(reused)
                    else:
                        Trace.log(Trace.EDGE_COND_FALSE, {
                            'nodes_to': edge['to'],
                            'nodes_from': edge['from'],
                            'condition': edge['condition_str'],
                            'flow_name': self._flow_name,
                            'parent': parent,
                            'node_args': self._node_args,
                            'dispatcher_id': self._dispatcher_id,
                            'selective': self._selective
                        })

            node_name = node['name']
            if not self._finished_nodes.get(node_name):
                self._finished_nodes[node_name] = []
            self._finished_nodes[node_name].append(node['id'])

        return new_started_nodes, selective_reuse

    def _start_and_update_retry(self):
        """Start the flow and update retry.

        :return: new/next retry
        """
        new_started_nodes = []
        selective_reuse = []

        Trace.log(Trace.FLOW_START, {'flow_name': self._flow_name,
                                     'dispatcher_id': self._dispatcher_id,
                                     'queue': Config.dispatcher_queues[self._flow_name],
                                     'selective': self._selective,
                                     'node_args': self._node_args})

        for i, start_edge in Config.get_starting_edges(self._flow_name):
            storage_pool = StoragePool(self._parent, self._flow_name)
            if start_edge['condition'](storage_pool, self._node_args):
                records, reused = self._fire_edge(i, start_edge, storage_pool,
                                                  node_args=self._node_args, parent=self._parent)
                self._triggered_edges_idx.add(i)
                new_started_nodes.extend(records)
                selective_reuse.extend(reused)
            else:
                Trace.log(Trace.EDGE_COND_FALSE, {
                    'nodes_to': start_edge['to'],
                    'nodes_from': start_edge['from'],
                    'condition': start_edge['condition_str'],
                    'flow_name': self._flow_name,
                    'parent': self._parent,
                    'node_args': self._node_args,
                    'dispatcher_id': self._dispatcher_id,
                    'selective': self._selective
                })

        # Update here so the strategy has correct new_started_nodes without nowait
        new_started_nodes = self._update_waiting_edges(new_started_nodes)
        return new_started_nodes, selective_reuse

    def _continue_and_update_retry(self, previously_reused):
        """Continue with the flow based on previous runs.

        :return: tuple describing newly started and reused nodes by selective run
        """
        new_started_nodes = []
        selective_reuse = []
        fallback_started = []

        new_finished, new_failed = self._get_successful_and_failed(previously_reused)

        if new_finished or new_failed:
            # Instantiate results lazily
            self._waiting_edges = self._idxs2items(Config.edge_table[self._flow_name], self._waiting_edges_idx)
            self._update_waiting_edges(new_finished)

            new_started_nodes, selective_reuse = self._start_new_from_finished(new_finished)

            # We wait until all active nodes finish if there are some failed nodes try to recover from failure,
            # otherwise mark flow as failed
            if not self._active_nodes and self._failed_nodes:
                fallback_started = self._compute_and_run_fallback()
                if not fallback_started and self._failed_nodes:
                    # We will propagate state information so we can correctly propagate parent and failed nodes in
                    # parent flow if there is any
                    # We use JSON in order to use result backend with JSON configured so objects are not pickled
                    state_dict = self.to_dict()
                    state_info = {
                        'finished_nodes': state_dict['finished_nodes'],
                        'failed_nodes': state_dict['failed_nodes']
                    }
                    raise FlowError(state_info)

        return new_started_nodes, selective_reuse, fallback_started

    def update(self):
        """Check the current state in the system and start new nodes if possible.

        :return: retry count - can be None (do not retry dispatcher) or time in seconds to retry
        """
        fallback_started = []

        if not self._active_nodes and not self._finished_nodes and not self._waiting_edges and not self._failed_nodes:
            # we are starting up
            started, reused = self._start_and_update_retry()
        else:
            started, reused, fallback_started = self._continue_and_update_retry([])

        while reused:
            # We do not need to retry if there are some tasks that we can continue with
            started, reused, fallback_started = self._continue_and_update_retry(reused)

        self._retry = Config.strategies[self._flow_name]({
            'previous_retry': self._retry,
            'active_nodes': self._active_nodes,
            'failed_nodes': self._failed_nodes,
            'new_started_nodes': started,
            'new_fallback_nodes': fallback_started,
            'finished_nodes': self._finished_nodes
        })

        return self._retry



#!/bin/env python3
# pragma: no cover

from functools import reduce


def argsFieldLessEqual(node_args, key, value):
    try:
        val = reduce(lambda m, k: m[k], key if isinstance(key, list) else [key], node_args)
        return val <= value
    except:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
# pylint: disable=line-too-long
"""Built-in tracing mechanism.

A list of events that can be traced:


+----------------------------+-------------------------------------+-----------------+------------------------------------+
| Event                      |  Event description                  | Emitter         |  msg_dict.keys()                   |
+============================+=====================================+=================+====================================+
|                            | A flow has been scheduled.          |                 |  dispatcher_id, selective,         |
|   `FLOW_SCHEDULE`          |                                     | Msg producer    |  queue, node_args,                 |
|                            |                                     |                 |  flow_name                         |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Dispatcher was started and will     |                 |  dispatcher_id, state, selective,  |
|   `DISPATCHER_WAKEUP`      | check flow status.                  | Dispatcher      |  retry, queue, node_args,          |
|                            |                                     |                 |  flow_name                         |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | A new flow is starting, this event  |                 |  dispatcher_id, selective,         |
|   `FLOW_START`             | is emitted on after                 | Dispatcher      |  queue, node_args, flow_name       |
|                            | `DISPATCHER_WAKEUP`.                |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Emitted when a new task is          |                 |  countdown, condition_str,         |
|   `TASK_SCHEDULE`          | scheduled by dispatcher.            | Dispatcher      |  task_name, queue,                 |
|                            |                                     |                 |  node_args, dispatcher_id, parent, |
|                            |                                     |                 |  task_id, selective, foreach_str,  |
|                            |                                     |                 |  flow_name, selective_edge         |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Emitted when a task is going to     |                 |                                    |
|   `TASK_START`             | be executed on worker side.         | Task            |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Emitted when a flow is scheduled    |                 |                                    |
|   `SUBFLOW_SCHEDULE`       | by dispatcher.                      | Dispatcher      |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalizing end of task execution,  |                 |                                    |
|   `TASK_END`               | task finished successfully.         | Task            |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalizing that a node in flow     |                 |                                    |
|   `NODE_SUCCESSFUL`        | graph (task or flow) successfully   | Dispatcher      |                                    |
|                            | finished.                           |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalizing that a task returned a  |                 |                                    |
|   `TASK_DISCARD_RESULT`    | value other than None, but no       | Task            |                                    |
|                            | storage was assigned to task to     |                 |                                    |
|                            | store result.                       |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalizing end of task, task       |                 |                                    |
|   `TASK_FAILURE`           | raised an exception, marking task   | Task            |                                    |
|                            | as failed.                          |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalizing that a task failed by,  |                 |                                    |
|   `TASK_RETRY`             | raising an exception and will be    | Task            |                                    |
|                            | retried.                            |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalization of flow failure,      |                 |                                    |
|   `FLOW_FAILURE`           | one more nodes in flow graph failed | Dispatcher      |                                    |
|                            | without successful fallback run.    |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | This event shouldn't be normally    |                 |                                    |
|   `DISPATCHER_FAILURE`     | seen - signalizing error in         | Dispatcher      |                                    |
|                            | Selinon.                            |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalization of captured failure   |                 |                                    |
|   `NODE_FAILURE`           | of node in task flow graph - flow   | Dispatcher      |                                    |
|                            | or task failure.                    |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Fallback for node failures in flow  |                 |                                    |
|   `FALLBACK_START`         | is started to handle node failures. | Dispatcher      |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Dispatcher finished scheduling new  |                 | dispatcher_id, state, selective,   |
|   `DISPATCHER_RETRY`       | nodes and will retry to check flow  | Dispatcher      | retry, queue, node_args,           |
|                            | status after a while.               |                 | state_dict, flow_name, parent      |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Flow has successfully ended.        |                 |                                    |
|   `FLOW_END`               |                                     | Dispatcher      |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Flow has migration was done.        |                 |                                    |
|   `MIGRATION`              |                                     | Dispatcher      |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalization of tainted flow when  |                 |                                    |
|   `MIGRATION_TAINTED_FLOW` | migration was run.                  | Dispatcher      |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Flow migration has error - skewed   |                 |                                    |
|   `MIGRATION_SKEW`         | migration.                          | Dispatcher      |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Given node did not connect to       |                 | storage_name                       |
|   `STORAGE_CONNECT`        | storage yet so a new connection is  | Dispatcher/Task |                                    |
|                            | requested by calling                |                 |                                    |
|                            | DataStorage.connect()               |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Called when disconnecting from      |                 |                                    |
|   `STORAGE_DISCONNECT`     | storage by calling                  | storage adapter |                                    |
|                            | DataStorage.disconnect()            | destructor      |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Retrieve task result from storage   |                 |                                    |
|   `STORAGE_RETRIEVE`       | that was assigned to the task.      | Dispatcher/Task |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Requested result of task was        |                 |                                    |
|   `STORAGE_RETRIEVED`      | retrieved.                          | Dispatcher/Task |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Store result of task in the         |                 |                                    |
|   `STORAGE_STORE`          | assigned storage.                   | Task            |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | The result of task has been stored  |                 |                                    |
|   `STORAGE_STORED`         | in the assigned storage.            | Task            |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | The condition on edge was evaluated |                 |                                    |
|   `EDGE_COND_FALSE`        | as false so destination nodes will  | Dispatcher      |                                    |
|                            | not be scheduled.                   |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Reports result of foreach function, |                 |                                    |
|   `FOREACH_RESULT`         | based on which N nodes will be      | Dispatcher      |                                    |
|                            | scheduled (N is runtime variable)   |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Requested state of the node (if the |                 |                                    |
|   `NODE_STATE_CACHE_GET`   | node was successful or failed) from | Dispatcher      |                                    |
|                            | cache.                              |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Requested entry in the state cache  |                 |                                    |
|   `NODE_STATE_CACHE_ADD`   | was not found and the node          | Dispatcher      |                                    |
|                            | succeeded so new entry to state     |                 |                                    |
|                            | is added.                           |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Requested entry was not found in    |                 |                                    |
|   `NODE_STATE_CACHE_MISS`  | state cache.                        | Dispatcher      |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Requested entry was found in the    |                 |                                    |
|   `NODE_STATE_CACHE_HIT`   | state cache and will be used.       | Dispatcher      |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Try to hit result cache for cached  |                 |                                    |
|   `TASK_RESULT_CACHE_GET`  | already requested task result.      | Dispatcher/Task |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Add to result cache result of a     |                 |                                    |
|   `TASK_RESULT_CACHE_ADD`  | task that was requested.            | Dispatcher/Task |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Requested task result was not found |                 |                                    |
|   `TASK_RESULT_CACHE_MISS` | in the result cache.                | Dispatcher/Task |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Requested task result was found     |                 |                                    |
|   `TASK_RESULT_CACHE_HIT`  | in the result cache.                | Dispatcher/Task |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Given edge will not be fired        |                 |                                    |
|   `SELECTIVE_OMIT_EDGE`    | as it is not part of direct path    | Dispatcher      |                                    |
|                            | to requested tasks on selective run.|                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | The desired node will not be        |                 |                                    |
|   `SELECTIVE_OMIT_NODE`    | scheduled as it is not requested    | Dispatcher      |                                    |
|                            | task nor dependency of requested    |                 |                                    |
|                            | task in selective task runs.        |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Keeps track of results of selective |                 | dispatcher_id, parent, selective,  |
|   `SELECTIVE_RUN_FUNC`     | run function on selective task      | Dispatcher      | node_args, flow_name,              |
|                            | runs.                               |                 | result, node_name                  |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalizes reuse of already         |                 |                                    |
|   `SELECTIVE_TASK_REUSE`   | computed results in selective task  | Dispatcher      |                                    |
|                            | runs.                               |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalizes storing error information|                 |                                    |
|   `STORAGE_STORE_ERROR`    | in assigned storage on task failure.| Task            |                                    |
|                            |                                     |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+
|                            | Signalizes that storing error       |                 |                                    |
| `STORAGE_OMIT_STORE_ERROR` | will not be done - missing storage  | Task            |                                    |
|                            | adapter or `store_error()` is not   |                 |                                    |
|                            | implemented.                        |                 |                                    |
+----------------------------+-------------------------------------+-----------------+------------------------------------+

"""

import datetime
import json
import logging
import os
import platform
import sys

_LOGGER = logging.getLogger(__name__)


class Trace:
    """Trace system flow actions."""

    _trace_functions = []
    _logger = None

    DISPATCHER_WAKEUP, \
        FLOW_SCHEDULE, \
        FLOW_START, \
        TASK_SCHEDULE, \
        TASK_START, \
        SUBFLOW_SCHEDULE, \
        TASK_END, \
        NODE_SUCCESSFUL, \
        TASK_DISCARD_RESULT, \
        TASK_FAILURE, \
        TASK_RETRY, \
        FLOW_FAILURE, \
        DISPATCHER_FAILURE, \
        NODE_FAILURE, \
        FALLBACK_START, \
        DISPATCHER_RETRY, \
        FLOW_END, \
        STORAGE_CONNECT, \
        STORAGE_DISCONNECT, \
        STORAGE_RETRIEVE, \
        STORAGE_RETRIEVED, \
        STORAGE_STORE, \
        STORAGE_STORED, \
        EDGE_COND_FALSE, \
        FOREACH_RESULT,\
        NODE_STATE_CACHE_GET,\
        NODE_STATE_CACHE_ADD,\
        NODE_STATE_CACHE_MISS,\
        NODE_STATE_CACHE_HIT,\
        TASK_RESULT_CACHE_GET,\
        TASK_RESULT_CACHE_ADD,\
        TASK_RESULT_CACHE_MISS,\
        TASK_RESULT_CACHE_HIT,\
        SELECTIVE_OMIT_EDGE,\
        SELECTIVE_OMIT_NODE,\
        SELECTIVE_RUN_FUNC,\
        SELECTIVE_TASK_REUSE,\
        STORAGE_STORE_ERROR,\
        STORAGE_OMIT_STORE_ERROR,\
        FALLBACK_COND_FALSE,\
        FALLBACK_COND_TRUE,\
        NODE_STATE_CACHE_ISSUE,\
        TASK_RESULT_CACHE_ISSUE,\
        STORAGE_ISSUE,\
        RESULT_BACKEND_ISSUE,\
        FLOW_RETRY,\
        MIGRATION,\
        MIGRATION_SKEW,\
        MIGRATION_TAINTED_FLOW, \
        MIGRATION_ERROR, \
        EAGER_FAILURE,\
        STORAGE_DELETE, \
        STORAGE_DELETED, \
        = range(53)

    WARN_EVENTS = (
        NODE_FAILURE,
        TASK_DISCARD_RESULT,
        TASK_RETRY,
        TASK_FAILURE,
        FLOW_FAILURE,
        STORAGE_OMIT_STORE_ERROR,
        NODE_STATE_CACHE_ISSUE,
        TASK_RESULT_CACHE_ISSUE,
        STORAGE_ISSUE,
        RESULT_BACKEND_ISSUE,
        FLOW_RETRY,
        MIGRATION_SKEW,
        MIGRATION_ERROR,
        EAGER_FAILURE,
    )

    _event_strings = (
        'DISPATCHER_WAKEUP',
        'FLOW_SCHEDULE',
        'FLOW_START',
        'TASK_SCHEDULE',
        'TASK_START',
        'SUBFLOW_SCHEDULE',
        'TASK_END',
        'NODE_SUCCESSFUL',
        'TASK_DISCARD_RESULT',
        'TASK_FAILURE',
        'TASK_RETRY',
        'FLOW_FAILURE',
        'DISPATCHER_FAILURE',
        'NODE_FAILURE',
        'FALLBACK_START',
        'DISPATCHER_RETRY',
        'FLOW_END',
        'STORAGE_CONNECT',
        'STORAGE_DISCONNECT',
        'STORAGE_RETRIEVE',
        'STORAGE_RETRIEVED',
        'STORAGE_STORE',
        'STORAGE_STORED',
        'EDGE_COND_FALSE',
        'FOREACH_RESULT',
        'NODE_STATE_CACHE_GET',
        'NODE_STATE_CACHE_ADD',
        'NODE_STATE_CACHE_MISS',
        'NODE_STATE_CACHE_HIT',
        'TASK_RESULT_CACHE_GET',
        'TASK_RESULT_CACHE_ADD',
        'TASK_RESULT_CACHE_MISS',
        'TASK_RESULT_CACHE_HIT',
        'SELECTIVE_OMIT_EDGE',
        'SELECTIVE_OMIT_NODE',
        'SELECTIVE_RUN_FUNC',
        'SELECTIVE_TASK_REUSE',
        'STORAGE_STORE_ERROR',
        'STORAGE_OMIT_STORE_ERROR',
        'FALLBACK_COND_FALSE',
        'FALLBACK_COND_TRUE',
        'NODE_STATE_CACHE_ISSUE',
        'TASK_RESULT_CACHE_ISSUE',
        'STORAGE_ISSUE',
        'RESULT_BACKEND_ISSUE',
        'FLOW_RETRY',
        'MIGRATION',
        'MIGRATION_SKEW',
        'MIGRATION_TAINTED_FLOW',
        'MIGRATION_ERROR',
        'EAGER_FAILURE',
        'STORAGE_DELETE',
        'STORAGE_DELETED'
    )

    def __init__(self):
        """Unused."""
        raise NotImplementedError()

    @classmethod
    def trace_by_logging(cls, logger=None):
        """Trace by using Python's logging.

        :param logger: optional logger that should be used
        """
        if not logger and not cls._logger:
            logger = logging.getLogger(__name__)

        cls._logger = logger
        cls._trace_functions.append(cls.logging_trace_func)

    @classmethod
    def trace_by_json(cls):
        """Trace by writing directly JSON trace points."""
        cls._trace_functions.append(cls.json_trace_func)

    @classmethod
    def trace_by_sentry(cls, dsn):
        """Trace using Sentry (see https://sentry.io).

        :param dsn: data source name for connecting to Sentry
        """
        try:
            import sentry_sdk
        except ImportError as exc:
            raise ImportError("Failed to import Sentry-SDK for Sentry logging, install it using `pip3 install sentry-sdk`")\
                from exc

        sentry_sdk.init(dsn.format(**os.environ))

    @classmethod
    def trace_by_func(cls, func):
        """Trace by a custom function.

        :param func: function with a one single argument
        """
        cls._trace_functions.append(func)

    @classmethod
    def log(cls, event, *msg_dict, **msg_dict_kwargs):
        """Log an event.

        :param event: tracing event
        :param msg_dict: message to be printed
        :param msg_dict_kwargs: kwargs like dictionary for traced details
        """
        if not cls._trace_functions:
            return

        to_report = {}
        for msg in msg_dict:
            to_report.update(msg)

        to_report.update(msg_dict_kwargs)

        for trace_func in cls._trace_functions:
            try:
                trace_func(event, to_report)
            except Exception as exc:  # pylint: disable=broad-except
                # To prevent from recursively reporting errors, this error is reported only to standard logging.
                _LOGGER.exception("Failed to report to trace function %r: %s", trace_func, str(exc))

    @classmethod
    def event2str(cls, event):
        """Translate event to it's string representation.

        :param event: event
        :return: string representation of event
        """
        return cls._event_strings[event]

    @classmethod
    def logging_trace_func(cls, event, msg_dict, logger=None):
        """Trace to Python's logging facilities.

        :param event: event that triggered trace point
        :param msg_dict: a dict holding additional trace information for event
        :param logger: a logger to be used, if None, logger from trace_by_logging will be used
        """
        report_dict = {
            'event': cls.event2str(event),
            'time': str(datetime.datetime.utcnow()),
            'details': msg_dict
        }
        message = "SELINON %10s - %s : %s" \
                  % (platform.node(), cls.event2str(event), json.dumps(report_dict, sort_keys=True))

        logger = logger or cls._logger
        if event == Trace.DISPATCHER_FAILURE:
            logger.error(message)
        elif event in cls.WARN_EVENTS:
            logger.warning(message)
        else:
            logger.info(message)

    @classmethod
    def json_trace_func(cls, event, msg_dict):
        """Trace by directly printing JSONs to stdout or stderr.

        :param event: event that triggered trace point
        :param msg_dict: a dict holding additional trace information for event
        """
        report_dict = {
            'event': cls.event2str(event),
            'time': str(datetime.datetime.utcnow()),
            'details': msg_dict,
            'node': platform.node()
        }
        # We could use simplejson here so this will be faster
        if event in cls.WARN_EVENTS:
            print(json.dumps(report_dict, sort_keys=True), file=sys.stderr)
        else:
            print(json.dumps(report_dict, sort_keys=True))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from flexmock import flexmock
from selinon_test_case import SelinonTestCase

from selinon.trace import Trace
from selinon import run_flow
from selinon.dispatcher import Dispatcher
from selinon import Retry
from selinon import UnknownFlowError


class TestOthers(SelinonTestCase):
    def test_trace_naming(self):
        for i, event_string in enumerate(Trace._event_strings):
            assert getattr(Trace, event_string) == i

    def test_run_flow(self):
        self.init(edge_table={'flow1': []}, dispatcher_queues={'flow1': 'flow1_queue'})

        flexmock(Dispatcher)
        Dispatcher.should_receive('apply_async').and_return("<dispatcher_id>")

        assert run_flow('flow1', node_args={'foo': 'bar'}) == "<dispatcher_id>"

    def test_run_unknown_flow_error(self):
        # We run uninitialized Config
        self.init(edge_table={}, dispatcher_queues=None)

        with pytest.raises(UnknownFlowError):
            run_flow('some_flow', node_args={'foo': 'bar'})

    def test_retry(self):
        countdown = 10
        assert Retry(countdown=countdown).countdown == countdown




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Selinon adapter for Amazon S3 storage."""

import json
import os

try:
    import boto3
    import botocore
except ImportError as exc:
    raise ImportError("Please install boto3 using `pip3 install selinon[s3]` in order to use S3 storage") from exc
from selinon import DataStorage, SelinonMissingDataException


class S3(DataStorage):
    """Amazon S3 storage adapter.

    For credentials configuration see boto3 library configuration
    https://github.com/boto/boto3
    """

    def __init__(self, bucket, location=None, endpoint_url=None, use_ssl=None,
                 aws_access_key_id=None, aws_secret_access_key=None, region_name=None, serialize_json=False):
        """Initialize S3 storage adapter from YAML configuration file.

        :param bucket: bucket name to be used
        :param location: AWS location
        :param endpoint_url: S3 endpoint (if local instance)
        :param use_ssl: True if SSL should be used
        :param aws_access_key_id: AWS access key
        :param aws_secret_access_key: AWS secret access key
        :param region_name: region to be used
        :param serialize_json: serialize JSON output (dict or list) to a blob - needed as S3 objects are blobs
        """
        # AWS access key and access id are handled by Boto - place them to config or use env variables
        super().__init__()
        self._bucket_name = bucket.format(**os.environ)
        self._location = location.format(**os.environ) if location else None
        self._s3 = None
        self._use_ssl = bool(use_ssl.format(**os.environ) if isinstance(use_ssl, str) else use_ssl)
        self._endpoint_url = endpoint_url.format(**os.environ) if endpoint_url else None
        self._serialize_json = serialize_json
        aws_access_key_id = aws_access_key_id.format(**os.environ) if aws_access_key_id else None
        aws_secret_access_key = aws_secret_access_key.format(**os.environ) if aws_secret_access_key else None
        region_name = region_name.format(**os.environ) if region_name else None
        self._session = boto3.session.Session(aws_access_key_id=aws_access_key_id,
                                              aws_secret_access_key=aws_secret_access_key,
                                              region_name=region_name)

    def is_connected(self):  # noqa
        return self._s3 is not None

    def connect(self):  # noqa
        # we need signature version v4 as new AWS regions use this version and we won't be able to connect without this
        self._s3 = self._session.resource('s3', config=botocore.client.Config(signature_version='s3v4'),
                                          use_ssl=self._use_ssl, endpoint_url=self._endpoint_url)

        # check that the bucket exists - see boto docs
        try:
            self._s3.meta.client.head_bucket(Bucket=self._bucket_name)
        except botocore.exceptions.ClientError as exc:
            # if a client error is thrown, then check that it was a 404 error.
            # if it was a 404 error, then the bucket does not exist.
            error_code = int(exc.response['Error']['Code'])
            if error_code == 404:
                self._s3.create_bucket(Bucket=self._bucket_name,
                                       CreateBucketConfiguration={
                                           'LocationConstraint': self._location
                                       })
            else:
                raise

    def disconnect(self):  # noqa
        if self._s3:
            del self._s3
            self._s3 = None

    def retrieve(self, flow_name, task_name, task_id):  # noqa
        assert self.is_connected()  # nosec

        blob = self._s3.Object(self._bucket_name, task_id).get()['Body'].read()
        if not self._serialize_json:
            return blob

        return json.loads(blob.decode())

    def store(self, node_args, flow_name, task_name, task_id, result):  # noqa
        assert self.is_connected()  # nosec

        if self._serialize_json:
            result = json.dumps(result).encode()

        self._s3.Object(self._bucket_name, task_id).put(Body=result)

    def store_error(self, node_args, flow_name, task_name, task_id, exc_info):  # noqa
        # just to make pylint happy
        raise NotImplementedError()

    def delete(self, flow_name, task_name, task_id):
        assert self.is_connected()  # nosec

        s3_object = self._s3.Object(self._bucket_name, task_id)

        try:
            s3_object.load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                # The object does not exist.
                raise SelinonMissingDataException from e
            raise e
        s3_object.delete()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

from selinon_test_case import SelinonTestCase

from selinon import SystemState


class TestNodeFailures(SelinonTestCase):
    def test_nowait_task(self):
        #
        # flow1:
        #
        #     Task1
        #
        # Note:
        #    Task1 is marked as nowait node
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true}],
            'flow2': []
        }
        nowait_nodes = {'flow1': ['Task1']}
        self.init(edge_table, nowait_nodes=nowait_nodes)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task1' not in state_dict.get('active_nodes')
        assert len(state_dict.get('waiting_edges')) == 0

    def test_nowait_flow(self):
        #
        # flow1:
        #
        #     flow2
        #
        # Note:
        #    flow2 is marked as nowait node
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['flow2'], 'condition': self.cond_true}],
            'flow2': []
        }
        nowait_nodes = {'flow1': ['flow2']}
        self.init(edge_table, nowait_nodes=nowait_nodes)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert 'flow2' in self.instantiated_flows
        assert 'flow2' not in state_dict.get('active_nodes')
        assert len(state_dict.get('waiting_edges')) == 0

    def test_nowait_in_flow(self):
        #
        # flow1:
        #
        #         Task1
        #           |
        #       ----------
        #      |          |
        #    Task2      Task3
        #
        # Note:
        #   Task3 finishes before Task2 and Task3 is marked as nowait
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        nowait_nodes = {'flow1': ['Task3']}
        self.init(edge_table, nowait_nodes=nowait_nodes)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 1

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, 1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 1
        assert 'Task3' not in state_dict.get('active_nodes')
        assert 'Task2' in state_dict['active_nodes'][0]['name']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import traceback
from .states import states


# TODO: rename finished_mapping
class AsyncResult:
    _finished_mapping = {}
    _result_mapping = {}

    def __init__(self, id):
        self._id = id

    def __repr__(self):
        return "<%s>" % self._id

    @classmethod
    def clear(cls):
        cls._finished_mapping = {}
        cls._result_mapping = {}

    @classmethod
    def finished_mapping(cls):
        return cls._finished_mapping

    @classmethod
    def set_finished(cls, task_id):
        cls._finished_mapping[task_id] = states.SUCCESS

    @classmethod
    def set_failed(cls, task_id):
        cls._finished_mapping[task_id] = states.FAILURE

    @classmethod
    def set_unfinished(cls, task_id):
        cls._finished_mapping[task_id] = states.STARTED

    @classmethod
    def set_result(cls, task_id, result):
        cls._result_mapping[task_id] = result

    def successful(self):
        return self._finished_mapping[self._id] == states.SUCCESS

    def failed(self):
        return self._finished_mapping[self._id] == states.FAILURE

    @property
    def result(self):
        return self._result_mapping[self._id]

    @property
    def traceback(self):
        assert self.failed()

        # there is no (documented) way on how to create Traceback object, so simulate it by rising exception
        # instance that was set
        try:
            raise self._result_mapping[self._id]
        except:
            return traceback.format_exc()

    @property
    def state(self):
        return self._finished_mapping[self._id]

    @property
    def status(self):
        return self._finished_mapping[self._id]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from selinon.errors import CacheMissError
from selinon.caches import (FIFO, LIFO, LRU, MRU, RR)
from selinon_test_case import SelinonTestCase

# Available caches that should be tested
_CACHE_TYPES = [
    FIFO,
    LIFO,
    LRU,
    MRU,
    RR
]


@pytest.mark.parametrize("cache_cls", _CACHE_TYPES)
class TestCache(SelinonTestCase):
    """
    Generic cache test cases
    """
    @staticmethod
    def _item_id2item(i):
        return "x%d" % i

    def test_empty_miss(self, cache_cls):
        cache = cache_cls(max_cache_size=3)

        with pytest.raises(CacheMissError):
            cache.get("item_id", "Task1", "flow1")

    def test_zero_items(self, cache_cls):
        cache = cache_cls(max_cache_size=0)

        with pytest.raises(CacheMissError):
            cache.get("item_id1", "Task1", "flow1")

        cache.add("item_id1", "item", "Task1", "flow1")

        with pytest.raises(CacheMissError):
            assert cache.get("item_id1", "Task1", "flow1") == "item"

    def test_one_item(self, cache_cls):
        cache = cache_cls(max_cache_size=1)

        cache.add("item_id1", "item", "Task1", "flow1")
        assert cache.get("item_id1", "Task1", "flow1") == "item"

    def test_two_items(self, cache_cls):
        cache = cache_cls(max_cache_size=2)

        cache.add("item_id1", 1, "Task1", "flow1")
        assert cache.get("item_id1", "Task1", "flow1") == 1

        cache.add("item_id2", 2, "Task1", "flow1")
        assert cache.get("item_id2", "Task1", "flow1") == 2

    def test_multiple_items(self, cache_cls):
        item_count = 16
        cache = cache_cls(max_cache_size=item_count)

        for item_id in range(item_count):
            cache.add(item_id, self._item_id2item(item_id), "Task1", "flow1")
            assert cache.get(item_id, "Task1", "flow1") == self._item_id2item(item_id)

        for item_id in range(item_count - 1, -1, -1):
            assert cache.get(item_id, "Task1", "flow1") == self._item_id2item(item_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Simulate execution in a single CLI run.

Using Selinon executor is a good way to test you configuration and behaviour locally if you would like to save
some time when debugging or exploring all the possibilities that Selinon offers you. Keep in mind that running
Selinon locally was designed for development purposes and behaviour in general can (and in many cases will) vary.

The key idea behind executor is to simulate exchange of messages which is done by your Celery broker and Kombu under
the hood of Celery. Thus there are lazily created queues that are referenced by their names (see
selinon.executor.time_queue for implementation details). These queues hold messages prior to time on which they were
scheduled. Under the hood there is used a heap queue to optimize inserting to O(log(N)) in the worst case where N is
number messages currently in the queue. These queues are coupled into QueuePool (selinon.execute.queue_pool) which
encapsulates all queues, keeps their references, instantiates it lazily and provides concurrency safety.

In order to avoid starving, QueuePool keeps track of the queue ("a last used queue") which prevents from
starving messages that were scheduled for the same time (basically a simple round-robin). QueuePool looks for a message
that is not scheduled to the future and can be executed (to do so we get O(N) where N is number of queues being used).

All workers listen on all queues for now. This prevents from waiting on a message that would be never processed.

In order to understand how Executor works, you need to understand how Celery works. Please refer to Celery
documentation if you are a Celery-newbie.
"""

import copy
from datetime import datetime
from datetime import timedelta
import logging
import traceback

from selinon.celery import Task as CeleryTask
from selinon import Config
from selinon import run_flow
from selinon import run_flow_selective
from selinon.system_state import SystemState
from selinon import UnknownError
from selinon.global_config import GlobalConfig

from .celery_mocks import simulate_apply_async
from .celery_mocks import simulate_retry
from .celery_mocks import SimulateAsyncResult
from .celery_mocks import SimulateRetry
from .progress import Progress
from .queue_pool import QueuePool


class Executor:
    """Executor that executes Selinon run in a multi-process environment."""

    executor_queues = QueuePool()
    _logger = logging.getLogger(__name__)

    DEFAULT_SLEEP_TIME = 1
    DEFAULT_CONCURRENCY = 1

    def __init__(self, nodes_definition, flow_definitions,
                 concurrency=DEFAULT_CONCURRENCY, sleep_time=DEFAULT_SLEEP_TIME,
                 config_py=None, keep_config_py=False, show_progressbar=True):
        """Instantiate execute.

        :param nodes_definition: path to nodes.yaml file
        :type nodes_definition: str
        :param flow_definitions: a list of YAML files describing flows
        :type flow_definitions: list
        :param concurrency: executor concurrency
        :type concurrency: int
        :param sleep_time: number of seconds to wait before querying queue
        :type sleep_time: float
        :param config_py: a path to file where Python code configuration should be generated
        :type config_py: str
        :param keep_config_py: if true, do not delete generated config.py
        :type keep_config_py: bool
        :param show_progressbar: show progressbar on executor run
        :type show_progressbar: bool
        """
        Config.set_config_yaml(nodes_definition, flow_definitions,
                               config_py=config_py,
                               keep_config_py=keep_config_py)

        self.concurrency = concurrency
        self.sleep_time = sleep_time
        self.show_progressbar = show_progressbar

        if concurrency != 1:
            raise NotImplementedError("Concurrency is now unsupported")

    @staticmethod
    def _prepare():
        """Prepare Selinon for executor run."""
        # We need to assign a custom async result as we are not running Celery but our mocks instead
        SystemState._get_async_result = SimulateAsyncResult  # pylint: disable=protected-access
        # Overwrite used Celery functions so we do not rely on Celery logic at all
        CeleryTask.apply_async = simulate_apply_async
        CeleryTask.retry = simulate_retry

    def run(self, flow_name, node_args=None):
        """Run executor.

        :param flow_name: a flow name that should be run
        :param node_args: arguments for the flow
        """
        self._prepare()
        run_flow(flow_name, node_args)
        self._executor_run()

    def run_flow_selective(self, flow_name, task_names, node_args=None, follow_subflows=False, run_subsequent=False):
        """Run only desired tasks in a flow.

        :param flow_name: name of the flow that should be run
        :param task_names: name of the tasks that should be run
        :param node_args: arguments that should be supplied to flow
        :param follow_subflows: if True, subflows will be followed and checked for nodes to be run
        :param run_subsequent: trigger run of all tasks that depend on the desired task
        :return: dispatcher id that is scheduled to run desired selective task flow
        :raises selinon.errors.SelectiveNoPathError: there was no way found to the desired task in the flow
        """
        self._prepare()
        run_flow_selective(
            flow_name,
            task_names,
            node_args,
            follow_subflows,
            run_subsequent
        )
        self._executor_run()

    def _executor_run(self):
        """Perform task execution based on published message on queue."""
        while not self.executor_queues.is_empty():
            # TODO: concurrency
            self._logger.debug("new executor run")

            # Retrieve a task that can be run right now
            time, record = self.executor_queues.pop()
            task, celery_kwargs = record

            # we got a task with the lowest wait time - we need to wait if the task was scheduled in the future
            wait_time = (time - datetime.now()).total_seconds()
            Progress.sleep(wait_time=wait_time,
                           sleep_time=self.sleep_time,
                           info_text='Waiting for next task to process (%s seconds)... ' % round(wait_time, 3),
                           show_progressbar=self.show_progressbar and self.concurrency == 1)
            try:
                kwargs = celery_kwargs.get('kwargs')
                # remove additional metadata placed by Selinon when doing tracing
                kwargs.pop('meta', None)
                # Perform deep copy so any modification on task arguments does not affect arguments in queues.
                result = task.run(**copy.deepcopy(kwargs))

                # Dispatcher needs info about flow (JSON), but SelinonTaskEnvelope always returns None - we
                # need to keep track of success)
                SimulateAsyncResult.set_successful(task.request.id, result)
            except SimulateRetry as selinon_exc:
                if 'exc' in selinon_exc.celery_kwargs and selinon_exc.celery_kwargs.get('max_retries', 1) == 0:
                    # log only user exception as we do not want SimulateRetry in our exception traceback
                    user_exc = selinon_exc.celery_kwargs['exc']
                    user_exc_info = (user_exc, user_exc, user_exc.__traceback__)
                    self._logger.exception(str(user_exc), exc_info=user_exc_info)
                    SimulateAsyncResult.set_failed(task.request.id, traceback.format_exception(*user_exc_info))
                else:
                    # reschedule if there was an exception and we did not hit max_retries when doing retry
                    Executor.schedule(task, selinon_exc.celery_kwargs)
            except KeyboardInterrupt:  # pylint: disable=try-except-raise
                raise
            except Exception as exc:
                raise UnknownError("Ooooops! Congratulations! It looks like you've found a bug! Feel free to open an "
                                   "issue at https://github.com/selinon/selinon/issues") from exc

    @classmethod
    def schedule(cls, task, celery_kwargs):
        """Schedule a new task to be executed.

        :param task: task to be executed
        :type task: Dispatcher|SelinonTaskEnvelope
        :param celery_kwargs: arguments for the task - raw Celery arguments which also carry additional Selinon
                              arguments
        """
        cls._logger.debug("executor is scheduling %s - %s", task.__class__.__name__, celery_kwargs)
        cls.executor_queues.push(queue_name=celery_kwargs.get('queue', GlobalConfig.DEFAULT_CELERY_QUEUE),
                                 time=datetime.now() + timedelta(seconds=celery_kwargs.get('countdown') or 0),
                                 record=(task, celery_kwargs,))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

from selinon import SystemState
from selinon import DataStorage
from selinon_test_case import SelinonTestCase


class _MyStorage(DataStorage):
    result = None

    def __init__(self):
        pass

    def connect(self):
        pass

    def disconnect(self):
        pass

    def is_connected(self):
        # return True so we can test retrieve()
        return True

    def store(self, node_args, flow_name, task_name, task_id, result):
        pass

    def retrieve(self, flow_name, task_name, task_id):
        return self.result


class TestNodeArgs(SelinonTestCase):
    def test_task2task(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     Task2
        #
        # Note:
        #    Result of Task1 is propagated to Task2 as node_args
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table,
                  node_args_from_first={'flow1': True},
                  task2storage_mapping={'Task1': 'Storage1'},
                  storage_mapping={'Storage1': _MyStorage()})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        # Run without change at first
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert system_state.node_args is None
        assert retry is not None
        assert 'Task1' in self.instantiated_tasks

        # Task1 has finished
        task1 = self.get_task('Task1')
        task1_result = "propagated result of Task1"
        self.set_finished(task1, task1_result)
        _MyStorage.result = task1_result

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        system_state.to_dict()

        task2 = self.get_task('Task2')
        assert 'Task2' in self.instantiated_tasks
        assert task2.node_args == task1_result
        assert retry is not None

    def test_task2flow(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     flow2
        #
        # Note:
        #    Result of Task1 is not propagated to flow2 as even node_args_from_first is set - but propagate_node_args
        #    is not set
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}],
            'flow2': []
        }
        self.init(edge_table,
                  node_args_from_first={'flow1': True},
                  task2storage_mapping={'Task1': 'Storage1'},
                  storage_mapping={'Storage1': _MyStorage()})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks

        # Run without change at first
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert system_state.node_args is None
        assert retry is not None
        assert 'Task1' in self.instantiated_tasks

        # Task1 has finished
        task1 = self.get_task('Task1')
        task1_result = "propagated result of Task1"
        self.set_finished(task1, task1_result)
        _MyStorage.result = task1_result

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        flow2 = self.get_flow('flow2')
        assert 'flow2' in self.instantiated_flows
        assert flow2.node_args is None
        assert retry is not None

    def test_task2flow_propagate(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     flow2
        #
        # Note:
        #    Result of Task1 is not propagated to flow2 as even node_args_from_first is set - but propagate_node_args
        #    is not set
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}],
            'flow2': []
        }
        self.init(edge_table,
                  node_args_from_first={'flow1': True},
                  propagate_node_args={'flow1': True},
                  task2storage_mapping={'Task1': 'Storage1'},
                  storage_mapping={'Storage1': _MyStorage()})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks

        # Task1 has finished
        task1 = self.get_task('Task1')
        task1_result = "propagated result of Task1"
        self.set_finished(task1, task1_result)
        _MyStorage.result = task1_result

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        flow2 = self.get_flow('flow2')
        assert 'flow2' in self.instantiated_flows
        assert flow2.node_args == task1_result
        assert retry is not None

    def test_task2tasks(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #    --------
        #   |        |
        # Task2    Task3
        #
        # Note:
        #    Result of Task1 is propagated to flow1 as node_args
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table,
                  node_args_from_first={'flow1': True},
                  task2storage_mapping={'Task1': 'Storage1'},
                  storage_mapping={'Storage1': _MyStorage()})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks

        # Task1 has finished
        task1 = self.get_task('Task1')
        task1_result = "propagated result of Task1"
        self.set_finished(task1, task1_result)
        _MyStorage.result = task1_result

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        assert retry is not None
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks

        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')

        assert task2.node_args == task1_result
        assert task3.node_args == task1_result

    def test_recurse(self):
        #
        # flow1:
        #
        #     Task1 <----
        #       |       |
        #       |       |
        #       ---------
        #
        # Note:
        #    Result of Task1 is propagated to flow1 as node_args
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table,
                  node_args_from_first={'flow1': True},
                  task2storage_mapping={'Task1': 'Storage1'},
                  storage_mapping={'Storage1': _MyStorage()})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks

        # Task1 has finished
        task1_1 = self.get_task('Task1')
        task1_result = "propagated result of Task1"
        self.set_finished(task1_1, task1_result)
        _MyStorage.result = task1_result

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        assert 'Task1' in self.instantiated_tasks
        task1_2 = self.get_task('Task1')
        assert task1_2.node_args == task1_result
        assert retry is not None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################


class GetTaskInstance:
    """
    Task instance factory used in system state to instantiate tasks
    """
    _flow_instances = []
    _task_instances = []

    @property
    def task_instances(self):
        """
        :return: all task instances that were instantiated
        """
        return self._task_instances

    @property
    def tasks(self):
        """
        :return: task names of tasks that were instantiated
        """
        return [t.task_name for t in self._task_instances]

    @property
    def flows(self):
        """
        :return: flow names of flows that were instantiated
        """
        return [f.flow_name for f in self._flow_instances]

    def task_by_name(self, task_name):
        """
        :param task_name: task name to look up
        :return: all tasks that were started with name task_name
        """
        return [t for t in self._task_instances if t.task_name == task_name]

    def flow_by_name(self, flow_name):
        """
        :param flow_name: flow name to look up
        :return: all flows that were started with name flow_name
        """
        return [f for f in self._flow_instances if f.flow_name == flow_name]

    @classmethod
    def register_node(cls, node):
        """
        Register a node that was scheduled
        :param node: flow to be tracked
        """
        # a custom task is used, see ./test/celery/task.py
        from celery.result import AsyncResult
        from selinon.dispatcher import Dispatcher

        if isinstance(node, Dispatcher):
            cls._flow_instances.append(node)
        else:
            cls._task_instances.append(node)
        AsyncResult.set_unfinished(node.task_id)

    @classmethod
    def remove_all_flows_by_name(cls, flow_name):
        """ Remove all flows with the given name """
        cls._flow_instances = [f for f in cls._flow_instances if f.flow_name != flow_name]

    @classmethod
    def remove_all_tasks_by_name(cls, task_name):
        """ Remove all flows with the given name """
        cls._task_instances = [t for t in cls._task_instances if t.task_name != task_name]

    @classmethod
    def clear(cls):
        """
        Since we are using class attributes and methos, we need to clear each time there will be a test
        :return:
        """
        cls._flow_instances = []
        cls._task_instances = []




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""MongoDB database adapter."""

import os
from selinon.data_storage import SelinonMissingDataException

try:
    from pymongo import MongoClient
except ImportError as exc:
    raise ImportError("Please install dependencies using `pip3 install selinon[mongodb]` "
                      "in order to use MongoStorage") from exc
from selinon import DataStorage


class MongoDB(DataStorage):
    """MongoDB database adapter."""

    def __init__(self, db_name, collection_name, host=None, port=27017):
        """Instantiate MongoDB storage adapter.

        :param db_name: MongoDB database name
        :param collection_name: MongoDB collection name
        :param host: MongoDB host
        :param port: MongoDB port
        """
        super().__init__()
        self.client = None
        self.collection = None
        self.db = None  # pylint: disable=invalid-name
        self.host = (host or 'localhost').format(**os.environ)
        self.port = int(port.format(**os.environ) if isinstance(port, str) else port)
        self.db_name = db_name.format(**os.environ)
        self.collection_name = collection_name.format(**os.environ)

    def is_connected(self):  # noqa
        return self.client is not None

    def connect(self):  # noqa
        self.client = MongoClient(self.host, self.port)
        self.db = self.client[self.db_name]
        self.collection = self.db[self.collection_name]

    def disconnect(self):  # noqa
        if self.is_connected():
            self.client.close()
            self.client = None
            self.db = None
            self.collection = None

    def retrieve(self, flow_name, task_name, task_id):  # noqa
        assert self.is_connected()  # nosec

        filtering = {'_id': 0}
        cursor = self.collection.find({'task_id': task_id}, filtering)

        if cursor.count() > 1:
            raise ValueError("Multiple records with same task_id found")

        if not cursor:
            raise FileNotFoundError("Record not found in database")

        record = cursor[0]

        assert task_name == record['task_name']  # nosec
        return record.get('result')

    def store(self, node_args, flow_name, task_name, task_id, result):  # noqa
        assert self.is_connected()  # nosec

        record = {
            'flow_name': flow_name,
            'node_args': node_args,
            'task_name': task_name,
            'task_id': task_id,
            'result': result

        }

        self.collection.insert(record)

        # task_id is unique here
        return task_id

    def store_error(self, node_args, flow_name, task_name, task_id, exc_info):  # noqa
        # just to make pylint happy
        raise NotImplementedError()


    def delete(self, flow_name, task_name, task_id):  # noqa
        assert self.is_connected()  # nosec

        filtering = {'_id': 0}
        cursor = self.collection.find({'task_id': task_id}, filtering)

        cursor_count = cursor.count()
        if cursor_count > 1:
            raise ValueError("Multiple records with same task_id found")

        if cursor_count == 0:
            raise SelinonMissingDataException("Record not found in database")

        self.collection.delete_one({'task_id': task_id})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Selinon command line interface."""
# pylint: disable=invalid-name,too-many-locals

from functools import partial
import json
import logging
import os
import sys

import click

from rainbow_logging_handler import RainbowLoggingHandler
from selinon import RequestError
from selinon import selinon_version
from selinon import selinon_version_codename
from selinon.executor import Executor
from selinon.helpers import git_previous_version
from selinon.helpers import git_previous_version_file
from selinon.migrations import Migrator
from selinon.migrations import TaintedFlowStrategy
from selinon.system import System
from selinon.user_config import UserConfig

_logger = logging.getLogger(os.path.basename(__file__))

_DEFAULT_NODE_ARGS = None
_DEFAULT_CONCURRENCY = 1
_DEFAULT_TAINTED_FLOW_STRATEGY = 'IGNORE'
_DEFAULT_PLOT_OUTPUT_DIR = '.'
_DEFAULT_PLOT_FORMAT = 'svg'
_DEFAULT_SLEEP_TIME = 0.5


def _validate_sleep_time(ctx, _, value):  # pylint: disable=unused-argument
    """Validate executor sleep time."""
    if value is not None and value < 0.0:
        raise click.BadParameter("Sleep time has to be positive number")
    return value


def _setup_logging(verbose, no_color):
    """Set up Python logging based on verbosity level.

    :param verbose: verbosity level
    :param no_color: do not use colorized output
    """
    level = logging.WARNING
    if verbose == 1:
        level = logging.INFO
    elif verbose > 1:
        level = logging.DEBUG

    logger = logging.getLogger()
    logger.setLevel(level)

    if not no_color:
        formatter = logging.Formatter("%(process)d: [%(asctime)s] %(name)s %(funcName)s:%(lineno)d: %(message)s")
        # setup RainbowLoggingHandler
        handler = RainbowLoggingHandler(sys.stderr)
        handler.setFormatter(formatter)
        logger.addHandler(handler)


def _print_version(ctx, _, value):
    """Print Selinon version and exit."""
    if not value or ctx.resilient_parsing:
        return
    click.echo(selinon_version)
    ctx.exit()


def _expand_flow_definitions(ctx, _, paths):  # pylint: disable=unused-argument
    """Expand path to configuration files if directory was supplied."""
    flow_definitions = []
    for path in paths:
        if os.path.isfile(path):
            flow_definitions.append(path)
            continue

        for root, _, files in os.walk(path, followlinks=True):
            for file in files:
                file_path = os.path.join(root, file)
                if file.endswith(('.yaml', '.yml')) and not file.startswith('.'):
                    _logger.debug("Found flow configuration file %r", file_path)
                    flow_definitions.append(file_path)
                else:
                    _logger.debug("File %r not considered as flow configuration file", file_path)

    return flow_definitions


@click.group()
@click.pass_context
@click.option('-v', '--verbose', count=True,
              help="Be verbose about what's going on (can be supplied multiple times).")
@click.option('--version', is_flag=True, is_eager=True, callback=_print_version, expose_value=False,
              help="Print Selinon version and exit.")
@click.option('--no-color', is_flag=True,
              help="Suppress colorized logging output.")
def cli(ctx=None, verbose=0, no_color=True):
    """Selinon command line interface."""
    if ctx:
        ctx.auto_envvar_prefix = 'SELINON'
    _setup_logging(verbose, no_color)


@cli.command()
@click.option('-n', '--nodes-definition', metavar='NODES.yml',
              type=click.Path(exists=True, dir_okay=False, readable=True), required=True,
              envvar='SELINON_NODES_DEFINITION',
              help="Path to nodes definition file.")
@click.option('-f', '--flow-definitions', metavar='FLOW.yml',
              type=click.Path(exists=True, readable=True), multiple=True, required=True,
              callback=_expand_flow_definitions, envvar='SELINON_FLOW_DEFINITIONS',
              help="Path to flow definition file (can be supplied multiple times) or path to a directory containing "
                   "YAML files.")
@click.option('--flow-name', metavar='FLOW_NAME', required=True,
              help="Specify a flow that should be run by its name.")
@click.option('-a', '--node-args', metavar='NODE_ARGS', default=_DEFAULT_NODE_ARGS,
              help="Specify arguments that should be passed to the executed flow (default: %s)." % _DEFAULT_NODE_ARGS)
@click.option('--node-args-file', metavar='FILE',
              help="Specify arguments that should be passed to the flow by a file.")
@click.option('-j', '--node-args-json', is_flag=True,
              help="Flow arguments are JSON, parse string representation into a dict.")
@click.option('-c', '--concurrency', metavar='PROCESS_COUNT', type=click.IntRange(1, None),
              default=_DEFAULT_CONCURRENCY,
              help="Worker count - number of processes that serve tasks in parallel "
                   "(default: %d)." % _DEFAULT_CONCURRENCY)
@click.option('-s', '--sleep-time', metavar='SLEEP_TIME', type=click.FLOAT, default=_DEFAULT_SLEEP_TIME,
              callback=_validate_sleep_time,
              help="Accuracy for worker sleeping when a task is scheduled to future "
                   "(default: %f)." % _DEFAULT_SLEEP_TIME)
@click.option('--config-py', type=click.Path(dir_okay=False, writable=True),
              help="Path to file where should be generated config.py placed")
@click.option('--keep-config-py', is_flag=True,
              help="Do not remove generated config.py file after run.")
@click.option('--hide-progressbar', is_flag=True,
              help="Hide progressbar during execution.")
@click.option('--selective-task-names', metavar="TASK1,TASK2,..",
              help="A comma separated list of tasks to which path should be computed on selective flow run.")
@click.option('--selective-follow-subflows', is_flag=True,
              help="Follow sub-flows in a selective flow run when computing path.")
@click.option('--selective-run-subsequent', is_flag=True,
              help="Run subsequent tasks (based on flow graph) affected by selective flow run.")
def execute(nodes_definition, flow_definitions, flow_name,
            node_args=_DEFAULT_NODE_ARGS, node_args_file=None, node_args_json=False, concurrency=_DEFAULT_CONCURRENCY,
            sleep_time=_DEFAULT_SLEEP_TIME, config_py=None, keep_config_py=False, hide_progressbar=False,
            selective_task_names=None, selective_follow_subflows=False, selective_run_subsequent=False):
    """Execute flows based on YAML configuration in a CLI."""
    if node_args and node_args_file:
        raise RequestError("Node arguments could be specified by command line argument or a file, but not from both")

    node_args = node_args
    if node_args_file:
        with open(node_args_file, 'r') as f:
            node_args = f.read()

    if node_args_json:
        try:
            node_args = json.loads(node_args)
        except Exception as e:
            raise RequestError("Unable to parse JSON arguments: %s" % str(e)) from e

    executor = Executor(nodes_definition, flow_definitions,
                        concurrency=concurrency, sleep_time=sleep_time,
                        config_py=config_py, keep_config_py=keep_config_py,
                        show_progressbar=not hide_progressbar)

    if selective_task_names:
        executor.run_flow_selective(
            flow_name,
            selective_task_names.split(','),
            node_args,
            follow_subflows=selective_follow_subflows,
            run_subsequent=selective_run_subsequent
        )
    else:
        if selective_follow_subflows:
            raise RequestError("Option --selective-follow-subflows requires --selective-task-names set")
        if selective_run_subsequent:
            raise RequestError("Option --selective-run-subsequent requires --selective-task-names set")

        executor.run(flow_name, node_args)


@cli.command()
@click.option('-n', '--nodes-definition', metavar='NODES.yml',
              type=click.Path(exists=True, dir_okay=False, readable=True), required=True,
              envvar='SELINON_NODES_DEFINITION',
              help="Path to tasks definition file.")
@click.option('-f', '--flow-definitions', metavar='FLOW.yml',
              type=click.Path(exists=True, readable=True), multiple=True, required=True,
              callback=_expand_flow_definitions, envvar='SELINON_FLOW_DEFINITIONS',
              help="Path to flow definition file (can be supplied multiple times) or path to a directory containing "
                   "YAML files.")
@click.option('-N', '--old-nodes-definition', metavar='NODES.yml',
              type=click.Path(exists=True, dir_okay=False, readable=True),
              help="Path to old nodes definition file.")
@click.option('-F', '--old-flow-definitions', metavar='FLOW.yml',
              type=click.Path(exists=True, readable=True), multiple=True,
              callback=_expand_flow_definitions,
              help="Path to old flow definition file (can be supplied multiple times) or path to a "
                   "directory containing YAML files.")
@click.option('--no-meta', is_flag=True,
              help="Do not add metadata information to generated migration files.")
@click.option('-m', '--migration-dir', type=click.Path(exists=True, writable=True, file_okay=False),
              help="Path to a directory containing generated migrations, overrides migration dir specified in "
                   "the nodes definition file.")
@click.option('-g', '--git', 'use_git', is_flag=True,
              help="Use Git VCS for obtaining old flow configuration.")
@click.option('--no-check', is_flag=True,
              help="Do not check system for errors.")
@click.option('-t', '--tainted-flows', type=click.Choice(['IGNORE', 'RETRY', 'FAIL']),
              default=_DEFAULT_TAINTED_FLOW_STRATEGY,
              help="Define strategy for tainted flows (default: %s)." % _DEFAULT_TAINTED_FLOW_STRATEGY)
def migrate(nodes_definition, flow_definitions, old_nodes_definition=None, old_flow_definitions=None, no_meta=False,
            migration_dir=None, use_git=False, no_check=False, tainted_flows=_DEFAULT_TAINTED_FLOW_STRATEGY):
    """Perform migrations on old and new YAML configuration files in flow changes."""
    # pylint: disable=too-many-branches
    if int(not old_flow_definitions) + int(not old_nodes_definition) == 1:
        raise RequestError("Please provide all flow and nodes configuration files or use --git")

    use_old_files = bool(old_flow_definitions)
    usage_clash = int(use_old_files) + int(use_git)
    if usage_clash == 2:
        raise RequestError("Option --git is disjoint with explicit configuration file specification")

    if usage_clash == 0:
        raise RequestError("Use --git or explicit old configuration file specification in order "
                           "to access old config files")

    if use_git:
        # Compute version that directly precedes the current master - there is relevant change
        # in any of config files.
        git_hash, depth = git_previous_version(nodes_definition)
        for new_flow_definition_file in flow_definitions:
            new_git_hash, new_depth = git_previous_version(new_flow_definition_file)
            if new_depth < depth:
                git_hash = new_git_hash
                depth = new_depth

        _logger.debug("Using Git hash %r for old config files", git_hash)
        old_nodes_definition = git_previous_version_file(git_hash, nodes_definition)
        old_flow_definitions = list(map(partial(git_previous_version_file, git_hash), flow_definitions))
    else:
        old_nodes_definition = old_nodes_definition
        old_flow_definitions = old_flow_definitions

    try:
        if not no_check:
            try:
                System.from_files(nodes_definition, flow_definitions)
            except Exception as e:
                raise RequestError("There is an error in your new configuration files: {}".format(str(e))) from e

            try:
                System.from_files(old_nodes_definition, old_flow_definitions)
            except Exception as e:
                raise RequestError("There is an error in your old configuration files: {}".format(str(e))) from e

        migrator = Migrator(migration_dir)
        new_migration_file = migrator.create_migration_file(
            old_nodes_definition,
            old_flow_definitions,
            nodes_definition,
            flow_definitions,
            TaintedFlowStrategy.get_option_by_name(tainted_flows),
            not no_meta
        )
    finally:
        if use_git:
            _logger.debug("Removing temporary files")
            # Clean up temporary files produced by git_previous_version()
            os.remove(old_nodes_definition)
            for old_flow_definition_file in old_flow_definitions:
                os.remove(old_flow_definition_file)

    _logger.info("New migration file placed in %r", new_migration_file)


@cli.command()
@click.option('-n', '--nodes-definition', metavar='NODES.yml',
              type=click.Path(exists=True, dir_okay=False, readable=True), required=True,
              envvar='SELINON_NODES_DEFINITION',
              help="Path to tasks definition file.")
@click.option('-f', '--flow-definitions', metavar='FLOW.yml',
              type=click.Path(exists=True, readable=True), multiple=True, required=True,
              callback=_expand_flow_definitions, envvar='SELINON_FLOW_DEFINITIONS',
              help="Path to flow definition file (can be supplied multiple times) or path to a directory containing "
                   "YAML files.")
@click.option('-c', '--config', metavar='CONFIG.yml', type=click.Path(dir_okay=False, exists=True),
              help="Path to a user configuration file containing graph styles.")
@click.option('-o', '--output-dir', default=_DEFAULT_PLOT_OUTPUT_DIR,
              help="Specify output dir where plotted graphs should be placed (default: %s)." % _DEFAULT_PLOT_OUTPUT_DIR)
@click.option('--format', 'image_format',
              help="Format of the output image (default: %s)." % _DEFAULT_PLOT_FORMAT)
def plot(nodes_definition, flow_definitions, config=None, output_dir=None, image_format=None):
    """Plot graphs of flows based on YAML configuration."""
    image_format = image_format or _DEFAULT_PLOT_FORMAT
    UserConfig.set_config(config)
    system = System.from_files(nodes_definition, flow_definitions)
    system.plot_graph(output_dir, image_format)


@cli.command()
@click.option('-n', '--nodes-definition', metavar='NODES.yml',
              type=click.Path(exists=True, dir_okay=False, readable=True), required=True,
              envvar='SELINON_NODES_DEFINITION',
              help="Path to tasks definition file.")
@click.option('-f', '--flow-definitions', metavar='FLOW.yml',
              type=click.Path(exists=True, readable=True), multiple=True, required=True,
              callback=_expand_flow_definitions, envvar='SELINON_FLOW_DEFINITIONS',
              help="Path to flow definition file (can be supplied multiple times) or path to a directory containing "
                   "YAML files.")
@click.option('-d', '--dump', metavar='DUMP.py', type=click.Path(dir_okay=False, writable=True),
              help="Generate Python code of the system.")
@click.option('--no-check', is_flag=True,
              help="Do not check system for errors.")
@click.option('--list-task-queues', is_flag=True,
              help="List all task queues based on task names.")
@click.option('--list-dispatcher-queues', is_flag=True,
              help="List dispatcher queues based on flow names.")
def inspect(nodes_definition, flow_definitions, dump=None, no_check=False,
            list_task_queues=False, list_dispatcher_queues=False):
    """Inspect Selinon configuration."""
    system = System.from_files(nodes_definition, flow_definitions, no_check)
    some_work = False

    if list_task_queues:
        for task_name, queue_name in system.task_queue_names().items():
            click.echo('%s:%s' % (task_name, queue_name))
        some_work = True

    if list_dispatcher_queues:
        for flow_name, queue_name in system.dispatcher_queue_names().items():
            click.echo('%s:%s' % (flow_name, queue_name))
        some_work = True

    if dump:
        system.dump2file(dump)
        some_work = True

    if not some_work:
        click.echo("Your configuration looks OK.")


@cli.command()
@click.option('--codename', is_flag=True,
              help="Get release codename.")
def version(codename=False):
    """Get version information."""
    try:
        from celery import __version__ as celery_version
    except ImportError:
        celery_version = 'Not Installed'

    if codename:
        click.echo(selinon_version_codename)
    else:
        click.echo('Selinon version: %s' % selinon_version)
        click.echo('Celery version: %s' % celery_version)


if __name__ == '__main__':
    cli()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""All user configurations generated from YAML file."""

import io
import logging
import os
import runpy
import tempfile

import celery

from .errors import ConfigNotInitializedError
from .errors import ConfigurationError
from .errors import UnknownStorageError
from .errors import UnsupportedCeleryError
from .system import System
from .trace import Trace


def requires_initialization(func):
    """Check that method that requires config can access initialized configuration."""
    def wrapper(class_, *args, **kwargs):
        """Wrap call for checking initialization."""
        if not class_.initialized:
            raise ConfigNotInitializedError("Selinon not initialized, cannot access configuration attributes")

        return func(class_, *args, **kwargs)

    return wrapper


class Config:
    """All user configurations generated from YAML file."""

    _logger = logging.getLogger(__name__)

    celery_app = None

    flows = {}
    task_classes = {}
    edge_table = {}
    nowait_nodes = {}
    eager_failures = None
    max_retry = None
    retry_countdown = None
    storage2storage_cache = {}
    storage_readonly = {}
    storage_task_name = {}
    propagate_node_args = {}
    propagate_parent = {}
    propagate_finished = {}
    propagate_compound_finished = {}
    output_schemas = None
    async_result_cache = {}
    migration_dir = None

    storage_mapping = {}
    task2storage_mapping = {}
    dispatcher_queues = {}
    task_queues = {}
    strategies = {}

    # Called from generated python code to mark that the configuration was correctly set up
    initialized = False

    @classmethod
    def _set_config(cls, config_module):
        """Set configuration from Python's module.

        :param config_module: config module to be used for attributes assignment
        """
        cls.task_classes = config_module['task_classes']
        cls.edge_table = config_module['edge_table']
        cls.failures = config_module['failures']
        cls.nowait_nodes = config_module['nowait_nodes']
        cls.eager_failures = config_module['eager_failures']
        cls.flows = list(cls.edge_table.keys())

        # misc
        cls.node_args_from_first = config_module['node_args_from_first']
        cls.async_result_cache = config_module['async_result_cache']

        # propagate_* entries
        cls.propagate_finished = config_module['propagate_finished']
        cls.propagate_node_args = config_module['propagate_node_args']
        cls.propagate_parent = config_module['propagate_parent']
        cls.propagate_compound_finished = config_module['propagate_compound_finished']

        # task configuration
        cls.output_schemas = config_module['output_schemas']
        cls.storage_mapping = config_module['storage2instance_mapping']
        cls.storage_task_name = config_module['storage_task_name']
        cls.task2storage_mapping = config_module['task2storage_mapping']
        cls.max_retry = config_module['max_retry']
        cls.retry_countdown = config_module['retry_countdown']
        cls.storage_readonly = config_module['storage_readonly']
        cls.storage2storage_cache = config_module['storage2storage_cache']

        # throttle configuration
        cls.throttle_tasks = config_module['throttle_tasks']
        cls.throttle_flows = config_module['throttle_flows']

        # queues
        cls.dispatcher_queues = config_module['dispatcher_queues']
        cls.task_queues = config_module['task_queues']

        # Dispatcher scheduling strategy
        cls.strategies = config_module['strategies']

        # Selective task configuration
        cls.selective_run_task = config_module['selective_run_task']

        # Configuration migrations
        cls.migration_dir = config_module['migration_dir']

        # call config init with Config class to set up other configuration specific values
        config_module['init'](cls)

    @classmethod
    def set_config_py(cls, config_code):
        """Set dispatcher configuration by Python config file.

        :param config_code: configuration source code
        """
        cls._logger.debug("Using config.py file from '%s'", config_code)
        config_module = runpy.run_path(config_code)
        cls._set_config(config_module)

    @classmethod
    def set_config_yaml(cls, nodes_definition_file, flow_definition_files, config_py=None, keep_config_py=False):
        """Set dispatcher configuration by path to YAML configuration files.

        :param nodes_definition_file: definition of system nodes - YAML configuration
        :param flow_definition_files: list of flow definition files
        :param config_py: a file that should be used for storing generated config.py
        :param keep_config_py: do not remove config_py file after run
        """
        system = System.from_files(nodes_definition_file, flow_definition_files)

        if not config_py:
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp_f:
                cls._logger.debug("Generating config.py file to created temporary file '%s'", tmp_f.name)
                system.dump2stream(tmp_f)
            config_py = tmp_f.name
        else:
            cls._logger.debug("Generating config.py file to proposed file '%s'", config_py)
            with open(config_py, "w") as output_f:
                system.dump2stream(output_f)

        cls.set_config_py(config_py)

        if not keep_config_py:
            cls._logger.debug("Removing generated config.py file '%s'", config_py)
            os.unlink(config_py)

    @classmethod
    def set_config_dict(cls, nodes_definition, flow_definitions):
        """Set configuration using dictionaries, no files are written to filesystem.

        :param nodes_definition: definition of nodes in the system
        :type nodes_definition: dict
        :param flow_definitions: a list with flow defintions
        :type flow_definitions: list
        """
        system = System.from_dict(nodes_definition, flow_definitions)
        dump = io.StringIO()
        system.dump2stream(dump)
        conf = {}
        # Ignore B102
        exec(dump.getvalue(), conf)  # pylint: disable=exec-used
        cls._set_config(conf)

    @classmethod
    def trace_by_func(cls, trace_func):
        """Set tracing function for Dispatcher.

        :param trace_func: a function that should be used to trace dispatcher actions
        """
        Trace.trace_by_func(trace_func)

    @classmethod
    def trace_by_logging(cls):
        """Use Python's logging for tracing."""
        Trace.trace_by_logging()

    @classmethod
    def trace_by_sentry(cls, dsn=None):
        """Use Sentry for tracing.

        :param dsn: DSN for sentry to be used (uses env variables if omitted, see Sentry docs)
        """
        Trace.trace_by_sentry(dsn)

    @classmethod
    def trace_by_json(cls):
        """Trace directly JSON output."""
        Trace.trace_by_json()

    @classmethod
    def set_celery_app(cls, celery_app):
        """Set celery application that should be used.

        :param celery_app: celery app instance
        """
        # Avoid circular imports
        from .dispatcher import Dispatcher
        from .task_envelope import SelinonTaskEnvelope

        cls._logger.debug("Registering Selinon to Celery context")

        cls.celery_app = celery_app

        celery_major_version = int(celery.__version__.split(".", maxsplit=1)[0])
        if celery_major_version == 4:
            celery_app.tasks.register(Dispatcher())
            celery_app.tasks.register(SelinonTaskEnvelope())
        elif celery_major_version == 5:
            celery_app.register_task(Dispatcher())
            celery_app.register_task(SelinonTaskEnvelope())
        else:
            raise UnsupportedCeleryError(
                "Unsupported Celery version {}, supported are celery>=4,<6".format(celery.__version__)
            )

    @classmethod
    def init(cls, celery_app, nodes_definition_file, flow_definition_files, config_py=None, keep_config_py=False):
        """Initialize Selinon configuration with Celery application.

        :param celery_app: celery application to be used
        :param nodes_definition_file: definition of system nodes - YAML configuration
        :param flow_definition_files: list of flow definition files
        :param config_py: a file that should be used for storing generated config.py
        :param keep_config_py: do not remove config_py file after run
        """
        # pylint: disable=too-many-arguments
        cls.set_config_yaml(nodes_definition_file, flow_definition_files, config_py, keep_config_py)
        cls.set_celery_app(celery_app)

    @staticmethod
    def _should_config(node_name, dst_node_name, configuration):
        """Syntax sugar for configuration entries that accept lists/booleans.

        :param node_name: node name
        :param dst_node_name: destination node to which configuration should be propagated
        :param configuration: configuration that should be checked
        :return: true if node_name satisfies configuration
        """
        if configuration[node_name] is True:
            return True

        if isinstance(configuration[node_name], list):
            return dst_node_name in configuration[node_name]

        return False

    @classmethod
    @requires_initialization
    def should_propagate_finished(cls, node_name, dst_node_name):
        """Check whether finished nodes should be propagated.

        :param node_name: node name that should be checked for propagate_finished
        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_finish
        """
        return cls._should_config(node_name, dst_node_name, cls.propagate_finished)

    @classmethod
    @requires_initialization
    def should_propagate_node_args(cls, node_name, dst_node_name):
        """Check whether node arguments should be propagated.

        :param node_name: node name that should be checked for propagate_node_args
        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_node_args
        """
        return cls._should_config(node_name, dst_node_name, cls.propagate_node_args)

    @classmethod
    @requires_initialization
    def should_propagate_parent(cls, node_name, dst_node_name):
        """Check whether parents be propagated.

        :param node_name: node name that should be checked for propagate_parent
        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_parent
        """
        return cls._should_config(node_name, dst_node_name, cls.propagate_parent)

    @classmethod
    @requires_initialization
    def should_propagate_compound_finished(cls, node_name, dst_node_name):  # pylint: disable=invalid-name
        """Check whether finished should be propagated (compound/flattered mode).

        :param node_name: node name that should be checked for propagate_compound_finished
        :param dst_node_name: destination node to which configuration should be propagated
        :return: True if should propagate_compound_finished
        """
        return cls._should_config(node_name, dst_node_name, cls.propagate_compound_finished)

    @classmethod
    @requires_initialization
    def get_task_instance(cls, task_name, flow_name, parent, task_id, dispatcher_id):
        """Get instance of SelinonTask.

        :param task_name: task name that should be instantiated (it is not necessarily SelinonTask)
        :param flow_name: flow name
        :param parent: parent nodes for instance
        :param task_id: id of the task
        :param dispatcher_id: id of dispatcher for flow flow
        :return: instance of the task
        """
        # pylint: disable=too-many-arguments
        task_class = Config.task_classes[task_name]
        return task_class(
            task_name=task_name,
            flow_name=flow_name,
            parent=parent,
            task_id=task_id,
            dispatcher_id=dispatcher_id
        )

    @classmethod
    @requires_initialization
    def is_flow(cls, node_name):
        """Check if given node is a flow by its name.

        :param node_name: name of the node to be checked
        :return: True if given node is a flow
        """
        return node_name in cls.flows

    @classmethod
    @requires_initialization
    def is_task(cls, node_name):
        """Check if given node is a task by its name.

        :param node_name: name of the node to be checked
        :return: True if given node is a task
        """
        return node_name in cls.task_classes

    @classmethod
    @requires_initialization
    def has_storage(cls, task_name):
        """Check whether the given task has assigned storage.

        :param task_name: name of a task
        :return: True if given task has assigned storage (either rw or readonly)
        """
        return task_name in cls.task2storage_mapping

    @classmethod
    @requires_initialization
    def has_readonly_storage(cls, task_name):
        """Check whether the given task has storage in read-only mode (results are not saved).

        :param task_name: name of a task
        :return: True if the given task has assigned readonly storage
        """
        return task_name in cls.storage_readonly

    @classmethod
    @requires_initialization
    def has_readwrite_storage(cls, task_name):
        """Check whether the given task has storage assigned and results are stored.

        :param task_name: name of a task
        :return: True if the given task has assigned rw storage
        """
        return cls.has_storage(task_name) and not cls.has_readonly_storage(task_name)

    @classmethod
    @requires_initialization
    def get_starting_edges(cls, flow_name):
        """Get starting edges for a flow.

        :param flow_name: a flow name to get starting edges
        :return: starting edges for a flow
        """
        if flow_name not in Config.edge_table:
            raise UnknownStorageError("No such flow in configuration: %s" % flow_name)

        start_edges = [(i, edge) for i, edge in enumerate(Config.edge_table[flow_name]) if len(edge['from']) == 0]
        if not start_edges:
            # This should not occur since selinon raises exception if such occurs, but just to be sure...
            raise ConfigurationError("No starting node found for flow '%s'!" % flow_name)

        return start_edges



#!/usr/bin/python3

import sys
import os
from setuptools import setup, find_packages


if sys.version_info[0] != 3:
    sys.exit("Python3 is required in order to install Selinon")


def get_requirements():
    with open('requirements.txt') as fd:
        return fd.read().splitlines()


def get_version():
    with open(os.path.join('selinon', 'version.py')) as f:
        version = f.readline()
    # dirty, remove trailing and leading chars
    return version.split(' = ')[1][1:-2]


def get_long_description():
    with open('README.rst', 'r') as f:
        return f.read()


setup(
    name='selinon',
    version=get_version(),
    entry_points={
        'console_scripts': ['selinon-cli=selinon.cli:cli']
    },
    packages=find_packages(),
    install_requires=get_requirements(),
    author='Fridolin Pokorny',
    author_email='fridolin.pokorny@gmail.com',
    maintainer='Fridolin Pokorny',
    maintainer_email='fridolin.pokorny@gmail.com',
    description='an advanced dynamic task flow management on top of Celery',
    long_description=get_long_description(),
    url='https://github.com/selinon/selinon',
    license='BSD',
    keywords='selinon celery yaml flow distributed-computing',
    extras_require={
        'celery': ['celery>=4,<6'],
        'mongodb': ['pymongo'],
        'postgresql': ['SQLAlchemy', 'SQLAlchemy-Utils'],
        'redis': ['redis'],
        's3': ['boto3'],
        'sentry': ['sentry-sdk']
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: System :: Distributed Computing",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy"
    ]
)
